% Copyright 2008 INRIA, F. Fages & J. Martin
% License GPL
%
% file eval.pl
%

:- use_module(library(terms)).
:- use_module(library(lists)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% built-ins partial evaluation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% evaluator (top-level)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
eval(node(L, Term), node(L, Eval)) :-
	!,
	eval(Term, Eval).
eval(Term, Eval) :-	
	Term =.. [Operator | Operands],
	(
	  (
	    is_compop(Term),
	    integers(Operands)
	  )
	;
	  is_arithop(Term)
	;
	  is_logop(Term)	
	;	  
	  is_listop(Term)	  
	),
	!,
	eval_rec(Operator, Operands, Eval).
eval(Term, Term).

% evaluator (term deconstruction)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% node functor
%%%%%%%%%%%%%%%%%%%%
eval_rec(F, [node(L, X)], node(L, Eval)) :-
	!,
	eval_rec(F, [X], Eval).
eval_rec(F, [node(LX, X), node(LY, Y)], Eval) :-
	!,
	eval_rec(F, [X, Y], Eval1),
	recompose_eval_term([F, X, Y], [LX, LY], Eval1, Eval).

% built-in functor
%%%%%%%%%%%%%%%%%%%%
eval_rec(F, [X], Eval) :-
	is_var(X),
	!,
	Eval =.. [F, X].
eval_rec(F, [X], Eval) :-
	!,
	eval_flat(F, X, Eval).
eval_rec(F, [X, Y], Eval) :-
	is_var(X),
	is_var(Y),
	!,
	Eval =.. [F, X, Y].
eval_rec(F, [X, Y], Eval) :-
	eval_flat(F, X, Y, Eval).

% arithmetic
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
eval_flat(+, 0, Y, Y) :-
	!.
eval_flat(+, X, 0, X) :-
	!.
eval_flat(+, X, Y, Z) :-
	integer(X),
	integer(Y),
	!,
	Z is X + Y.

eval_flat(-, 0, Y, Y) :-
	!.
eval_flat(-, X, 0, X) :-
	!.
eval_flat(-, X, Y, Z) :-
	integer(X),
	integer(Y),
	!,
	Z is X - Y.

eval_flat(*, 1, Y, Y) :-
	!.
eval_flat(*, X, 1, X) :-
	!.
eval_flat(*, 0, _, 0) :-
	!.
eval_flat(*, _, 0, 0) :-
	!.
eval_flat(*, X, Y, Z) :-
	integer(X),
	integer(Y),
	!,
	Z is X * Y.

eval_flat(/, X, 1, X) :-
	!.
eval_flat(/, 0, _, 0) :-
	!.
eval_flat(/, X, Y, Z) :-
	integer(X),
	integer(Y),
	!,
	Z is integer(X / Y).

eval_flat(min, min_integer, _, min_integer) :-
	!.
eval_flat(min, _, min_integer, min_integer) :-
	!.
eval_flat(min, max_integer, X, X) :-
	!.
eval_flat(min, X, max_integer, X) :-
	!.
eval_flat(min, X, Y, Z) :-
	integer(X),
	integer(Y),
	!,
	Z is min(X, Y).

eval_flat(max, max_integer, _, max_integer) :-
	!.
eval_flat(max, _, max_integer, max_integer) :-
	!.
eval_flat(max, min_integer, X, X) :-
	!.
eval_flat(max, X, min_integer, X) :-
	!.
eval_flat(max, X, Y, Z) :-
	integer(X),
	integer(Y),
	!,
	Z is max(X, Y).

eval_flat(log, B, X, Y) :-
	integer(X),
	!,
	Y is ceiling(log(B, X)).

eval_flat(exp, X, Y, Z) :-
	integer(X),
	integer(Y),
	!,
	Z is ceiling(exp(X, Y)).

% comparison
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
eval_flat(<, X, Y, 1) :-
	X < Y,
	!.

eval_flat(=<, X, Y, 1) :-
	X =< Y,
	!.

eval_flat(=, X, Y, 1) :-
	X =:= Y,
	!.

eval_flat(#, X, Y, 1) :-
	X =\= Y,
	!.

eval_flat(>=, X, Y, 1) :-
	X >= Y,
	!.

eval_flat(>, X, Y, 1) :-
	X > Y,
	!.

eval_flat(Op, _, _, 0) :-
	T =.. [Op, _, _],
	is_compop(T),
	!.

% logic
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
eval_flat(and, 1, Q, Q) :-
	!.
eval_flat(and, P, 1, P) :-
	!.
eval_flat(and, 0, _, 0) :-
	!.
eval_flat(and, _, 0, 0) :-
	!.

eval_flat(or, 1, _, 1) :-
	!.
eval_flat(or, _, 1, 1) :-
	!.
eval_flat(or, P, 0, P) :-
	!.
eval_flat(or, 0, Q, Q) :-
	!.

eval_flat(implies, 1, P, P) :-
	!.
eval_flat(implies, _, 1, 1) :-
	!.
eval_flat(implies, 0, _, 1) :-
	!.
eval_flat(implies, P, 0, not P) :-
	!.

eval_flat(equiv, 1, P, P) :-
	!.
eval_flat(equiv, P, 1, P) :-
	!.
eval_flat(equiv, 0, P, not P) :-
	!.
eval_flat(equiv, P, 0, not P) :-
	!.

% list
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
eval_flat(cons, H, T, [H | T]) :-
	!.

% set
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%eval_flat(\-\, S1, S2, S) :- !, S =-= S1 \-\ S2.
%eval_flat(/-\, S1, S2, S) :- !, S =-= S1 /-\ S2.
%eval_flat(\-/, S1, S2, S) :- !, S =-= S1 \-/ S2.

eval_flat(Op, P, Q, T) :-
	T =.. [Op, P, Q].

% unary ops evaluation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
eval_flat(not, 1, 0) :-
	!.
eval_flat(not, 0, 1) :-
	!.

eval_flat(abs, X, Z)  :-
	!,
	Z is abs(X).

%eval_flat(card, S, N) :- !, N =-= card(S).

eval_flat(C, P, T) :-
	T =.. [C, P].

% term reconstruction
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
recompose_eval_term(TermAsList, TermLabels, EvalTerm, REvalTerm) :-
	functor(EvalTerm, EFunctor, EArity),
	recompose_eval_term(TermAsList, TermLabels,
			    EvalTerm, EFunctor, EArity, REvalTerm).

recompose_eval_term([F, X, Y], [LX, LY], EvalTerm, F, 2, REvalTerm) :-
	arg(1, EvalTerm,  X),
	arg(2, EvalTerm,  Y),	
	!,
	REvalTerm =.. [F, node(LX, X), node(LY, Y)].
recompose_eval_term(_, _, EvalTerm, _, 2, EvalTerm) :-
	!.

recompose_eval_term([F, X, _Y], [LX, _LY], EvalTerm, F, 1, REvalTerm) :-
	arg(1, EvalTerm,  X),
	!,
	REvalTerm =.. [F, node(LX, X)].
recompose_eval_term([F, _X, Y], [_LX, LY], EvalTerm, F, 1, REvalTerm) :-
	arg(1, EvalTerm,  Y),
	!,
	REvalTerm =.. [F, node(LY, Y)].
recompose_eval_term(_, _, EvalTerm, _, 1, EvalTerm) :-
	!.

recompose_eval_term(_, _, EvalTerm, _, 0, EvalTerm).


% sets
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% :- use_module(library(sets)).

% S =-= S1 \-/ S2 :- is_set(S1), is_set(S2), union(S1,S2,S).
% S =-= S1 /-\ S2 :- is_set(S1), is_set(S2), intersection(S1,S2,S).
% S =-= S1 \-\ S2 :- is_set(S1), is_set(S2), subtract(S1,S2,S).
% N =-= card(S)   :- is_set(S), length(S,N).
% S =-= S         :- is_set(S).

% S =-= S1 \-/ S2 :- !, union(S1,S2,S).
% S =-= S1 /-\ S2 :- !, intersection(S1,S2,S).
% S =-= S1 \-\ S2 :- !, difference(S1,S2,S).
% N =-= card(S)   :- !, length(S,N).
% S =-= S.

% lt(X,Y) :- ((nonvar(X), nonvar(Y)) -> X<Y; true).

% union([],S,S) :- !.
% union(S,[],S) :- S\=[], !.
% union([X|TX],[X|TY],[X|TZ]):- !,
%    union(TX,TY,TZ).
% union([X|TX],[Y|TY],[X|TZ]):- !,
%    lt(X,Y),
%    union(TX,[Y|TY],TZ).
% union([X|TX],[Y|TY],[Y|TZ]):-
%    lt(Y,X),
%    union([X|TX],TY,TZ).
   
% intersection([],_,[]) :- !.
% intersection(S,[],[]) :- S\=[], !.
% intersection([X|TX],[X|TY],[X|TZ]):- !,
%    intersection(TX,TY,TZ).
% intersection([X|TX],[Y|TY],TZ):- !,
%    lt(X,Y),
%    intersection(TX,[Y|TY],TZ).
% intersection([X|TX],[Y|TY],TZ):-
%    lt(Y,X),
%    intersection([X|TX],TY,TZ).

% difference([],_,[]) :- !.
% difference(S,[],S):-S\=[], !.
% difference([X|TX],[X|TY],TZ):- !,
%    difference(TX,TY,TZ).
% difference([X|TX],[Y|TY],[X|TZ]):- !,
%    lt(X,Y),
%    difference(TX,[Y|TY],TZ).
% difference([X|TX],[Y|TY],TZ):-
%    lt(Y,X),
%    difference([X|TX],TY,TZ).

