% Copyright 2008 INRIA, F. Fages & J. Martin
% License GPL
%
% file order.pl
%

:- use_module(library(samsort)).

:- ensure_loaded(library(terms)).
:- ensure_loaded(library(lists)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% binary to n-ary and/or tree
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
naryfication_and_or(Tree, NaryTree) :-
	naryfication_and_or(Tree, [], [], NaryTree).

naryfication_and_or(Term, OStack, CStack, NTerm) :-
	conj_or_disj_op(Term, Op, P, Q),
	naryfication_and_or(P, [Op | OStack], CStack, P1),
	P1 =.. [Op, _-P1s],
	naryfication_and_or(Q, [Op | OStack], CStack, Q1),
	Q1 =.. [Op, _-Q1s],
	!,
	append(P1s, Q1s, L),
	NTerm =.. [Op, CStack-L].
naryfication_and_or(Term, OStack, CStack, NTerm) :-
	conj_or_disj_op(Term, Op, P, Q),
	naryfication_and_or(P, [Op | OStack], CStack, P1),
	P1 =.. [Op, _-P1s],
	naryfication_and_or(Q, [Op | OStack], CStack, Q1),
	!,
	append(P1s, [Q1], L),
	NTerm =.. [Op, CStack-L].
naryfication_and_or(Term, OStack, CStack, NTerm) :-
	conj_or_disj_op(Term, Op, P, Q),
	naryfication_and_or(P, [Op | OStack], CStack, P1),
	naryfication_and_or(Q, [Op | OStack], CStack, Q1),
	Q1 =.. [Op, _-Q1s],
	!,
	NTerm =.. [Op, CStack-[P1 | Q1s]].
naryfication_and_or(Term, OStack, CStack, NTerm) :-
	conj_or_disj_op(Term, Op, P, Q),
	naryfication_and_or(P, [Op | OStack], CStack, P1),
	naryfication_and_or(Q, [Op | OStack], CStack, Q1),
	!,
	NTerm =.. [Op, CStack-[P1, Q1]].
naryfication_and_or(node(L, P), [Op | OStack], CStack, NP) :- 
	naryfication_and_or(P, [Op | OStack], [Op-L | CStack], NP),
	!.
naryfication_and_or(P, _OSTack, CStack, node(CStack, P)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  n-ary to binary and/or tree
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
binaryfication_and_or(and(_-Conjuncts), _, BConjunct) :-
	binaryfication_and_or(Conjuncts, and, BConjunct),
	!.
binaryfication_and_or(or(_-Disjuncts), _, BDisjunct) :-
	binaryfication_and_or(Disjuncts, or, BDisjunct),
	!.
binaryfication_and_or(node(_L, P), Connective, BP) :-
	binaryfication_and_or(P, Connective, BP),
	!.
binaryfication_and_or([Xjunct], Connective, BXjunct) :-
	binaryfication_and_or(Xjunct, Connective, BXjunct),
	!.
binaryfication_and_or([Xjunct | Xjuncts], Connective, Bjunct) :- 
	binaryfication_and_or(Xjunct, Connective, BXjunct),
	binaryfication_and_or(Xjuncts, Connective, BXjuncts),
	!,
	Bjunct =.. [Connective, BXjunct, BXjuncts].
binaryfication_and_or(Xjunct, _, Xjunct).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% and/or tree ordering
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
order_and_or_tree(Tree, ConjCrits, DisjCrits, OrderedTree) :-
				%write('\"------------ \"'),nl,
				%write('-OrderingTree-\t\t'),nl, write(Tree),nl,
				%write('-ConjCrits-\t'), write(ConjCrits),nl,
				%write('-DisjCrits-\t'), write(DisjCrits),nl,
	crit_order_list(ConjCrits, ConjOrders),
	crit_order_list(DisjCrits, DisjOrders),
	naryfication_and_or(Tree, NaryTree),
				%write('-OrderingNaryTree-\t'), nl,write(NaryTree),nl,nl,
				%write('\"------------ \"'),nl,
	order_and_or_tree(NaryTree, none-1, [],
			  ConjOrders, ConjCrits, DisjOrders, DisjCrits,
			  _-(_-OrderedNaryTree)),
				%write('-OrderedNaryTree-\t'), nl, write(OrderedNaryTree),nl,nl,
	binaryfication_and_or(OrderedNaryTree, _, OrderedTree).

% and subtree
%%%%%%%%%%%%%%%%%%%%
order_and_or_tree(and(Labels-Conjuncts), _-N, CStack,
		  COrders, CCrits, DOrders, DCrits,
		  LexLabels-(N-and(Labels-OrderedConjuncts))) :-
	!,
	order_and_or_tree(Conjuncts, and-1, CStack,
			  COrders, CCrits, DOrders, DCrits,
			  LexConjuncts),
	lex_vector(Labels, DCrits, CStack, LexLabels),
	order_crit(1, COrders, [LexConjuncts], OrderedLexConjunct),
	filter_constraints(OrderedLexConjunct, OrderedConjuncts).

% or subtree
%%%%%%%%%%%%%%%%%%%%
order_and_or_tree(or(Labels-Disjuncts), _-N, CStack,
		  COrders, CCrits, DOrders, DCrits,
		  LexLabels-(N-or(Labels-OrderedDisjuncts))) :-
	!,
	order_and_or_tree(Disjuncts, or-1, CStack,
			  COrders, CCrits, DOrders, DCrits,
			  LexDisjuncts),
	lex_vector(Labels, CCrits, CStack, LexLabels),
	order_crit(1, DOrders, [LexDisjuncts], OrderedLexDisjunct),
	filter_constraints(OrderedLexDisjunct, OrderedDisjuncts).

% list of nodes
%%%%%%%%%%%%%%%%%%%%
order_and_or_tree([], _, _CStack,
		  _COrders, _CCrits, _DOrders, _DCrits,
		  []) :-
	!.
order_and_or_tree([P | Ps], Connective-N, CStack, 
		  COrders, CCrits, DOrders, DCrits,
		  [LexP | LexPs]) :-
	!,
	order_and_or_tree(P, Connective-N, CStack, 
			  COrders, CCrits, DOrders, DCrits,
			  LexP),
	M is N + 1,
	order_and_or_tree(Ps, Connective-M, CStack,
			  COrders, CCrits, DOrders, DCrits,
			  LexPs).

% conjunct node 
%%%%%%%%%%%%%%%%%%%%
order_and_or_tree(node(Labels, P), and-N, CStack,
		  COrders, CCrits, DOrders, DCrits,
		  LexLabel-(N-node(Labels, OP))) :-
	!,
	lex_vector(Labels, CCrits, CStack, LexLabel),
	order_and_or_tree(P, and-N, [Labels | CStack],
			  COrders, CCrits, DOrders, DCrits,
			  OP).

% disjunct node 
%%%%%%%%%%%%%%%%%%%%
order_and_or_tree(node(Labels, P), or-N, CStack,
		  COrders, CCrits, DOrders, DCrits,
		  LexLabel-(N-node(Labels, OP))) :-
	!,
	lex_vector(Labels, DCrits, CStack, LexLabel),	
	order_and_or_tree(P, or-N, [Labels | CStack], 
			  COrders, CCrits, DOrders, DCrits,
			  OP).

% conjunct leaf 
%%%%%%%%%%%%%%%%%%%%
order_and_or_tree(P, and-N, CStack, 
		  _COrders, CCrits, _DOrders, _DCrits,
		  LexVector-(N-P)) :-
	!,
	lex_vector(P, CCrits, CStack, LexVector).

% disjunct leaf 
%%%%%%%%%%%%%%%%%%%%
order_and_or_tree(P, or-N, CStack,
		  _COrders, _CCrits, _DOrders, DCrits,
		  LexVector-(N-P)) :-
	lex_vector(P, DCrits, CStack, LexVector).

% criteria values
%%%%%%%%%%%%%%%%%%%%
lex_vector(_P, [], _, []) :-
   !.
lex_vector([], _Crits, _CStack, []) :-
   !.
lex_vector([H | T], Crits, CStack, [LexComponentsH | LexComponents]) :-
   !,
   lex_vector(H, Crits, CStack, LexComponentsH),
   lex_vector(T, Crits, CStack, LexComponents).
lex_vector(P, [(MExpr if ^ is MTerm) | Crits], CStack, [LexComponent | LexComponents]) :-
   MExpr =.. [Order, Expr],
   functor(MTerm, Functor, Arity),
   functor(Pattern, Functor, Arity),
   (  
      matches(P, Pattern)
   ->
      MTerm =.. [Functor | FormalArgs],
      _-Q = P,
      Q =.. [Functor | Args],
      env_new(EmptyEnv),
      env_add_list(FormalArgs, EmptyEnv, Args, Env),
      substitution(Env, Expr, '', _, SubstExpr),
      rewrite_term(SubstExpr, CStack, Env, LexComponent)
   ;  
      (  
         Order == greatest
      -> 
         least_term(LexComponent)
      ;  
         greatest_term(LexComponent)
      )
   ),
   lex_vector(P, Crits, CStack, LexComponents).

matches(P, Pattern) :-
   P = Pattern.
matches(P, Pattern) :-
   P = _-Pattern.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% variable/value ordering
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% labeling terms (variable ordering)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
order_labelings(Labelings, OrderedLabelings) :-
	variables(Labelings, LabelingsVars),	
	(
	  (
	    varorder_terms(LexVarOrderCriteria),
	    LexVarOrderCriteria \== []
	  )
	->
				%write(variable_ordering),
				%write(LexVarOrderCriteria),nl,	  
	  order_variables(LabelingsVars, LexVarOrderCriteria, OrderedLabelingVars)
	;
	  OrderedLabelingVars = LabelingsVars
	),
				%write(ordered_variables),
				%write(OrderedLabelingVars),nl,
	(
	  (
	    valorder_terms(ValOrderCriteria),
	    ValOrderCriteria \== []
	  )
	->
				%write(value_ordering),
				%write(ValOrderCriteria),nl,
	  order_values(OrderedLabelingVars, ValOrderCriteria, OrderedLabelingVarsVal),
				%write(ordered_variables_values),
				%write(OrderedLabelingVarsVal),nl,
	  merge_labelings(OrderedLabelingVarsVal, OrderedLabelings)
	;
	  OrderedLabelings = labeling([], OrderedLabelingVars)
	).
				%,write(ordered_labelings),
				%write(OrderedLabelings),nl,nl.


% variables ordering
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
order_variables(Vars, LexVarOrderCriteria, OrderedVars) :-
				%write('-OrderingVars-\t\t'), write(Vars),nl,
				%write('-LexVarOrderCrit-\t'),write(LexVarOrderCriteria),nl,
	crit_order_list_0(LexVarOrderCriteria, OrderList),
				%write('-OrderList-\t\t'),write(OrderList),nl,
	lexico_terms(Vars, 1, LexVarOrderCriteria, LexVars),
				%write('-Lex vars-\t\t'),nl, write_list(LexVars),nl,
	order_crit(1, OrderList, [LexVars], OrderedLexTerm),
				%write('-OrderedLexTerm-\t'),nl, write_list(OrderedLexTerm),nl,	
	filter_variables(OrderedLexTerm, OrderedVars).

crit_order_list_0([], []).
crit_order_list_0([LexVarOrderCrit | LexVarOrderCrits], [Order | OrderList]) :-
   LexVarOrderCrit =.. [Order | _],
   crit_order_list_0(LexVarOrderCrits, OrderList).

crit_order_list([], []).
crit_order_list([LexVarOrderCrit | LexVarOrderCrits], [Order|OrderList]) :-
   %write(LexVarOrderCrit),nl,
   LexVarOrderCrit = (MExpr if ^ is _MTerm), 
   MExpr =.. [Order, _],
   crit_order_list(LexVarOrderCrits, OrderList).

lexico_terms([], _, _, []).
lexico_terms([Var|Vars], N, LexVarOrderCriteria, [LexTerm-(N-Var)|LexTerms]) :-
	Var = free_var(Name,_,_),
				%write(var(Var)),nl,
	M is N+1,
	lexico_term(Name, LexVarOrderCriteria, LexTerm),			
	lexico_terms(Vars, M, LexVarOrderCriteria, LexTerms).

lexico_term(_, [], []).
lexico_term(Name,
	    [LexVarOrderCriterion | LexVarOrderCriteria],
	    [LexComponent | LexComponents]) :-
				%write(Name-LexVarOrderCriterion),nl,
	subst(^, Name, LexVarOrderCriterion, TermSubst),
	TermSubst =.. [Order , TermToParse],        
	parse_expr(TermToParse, TermToRewrite),
				%write(Order-termtorewrite(TermToRewrite)),nl,
	on_exception(
		     _Error,
		     (
		       env_new(Env),                      
		       rewrite_term(TermToRewrite, [], Env, LexCompnt)
		     ),
		     (
		       (
			 Order == greatest
		       ->			 
			 least_term(LexCompnt) 
		       ;
			 greatest_term(LexCompnt) 
		       )
		     )
		    ),
       	variables(LexCompnt, LexVars),
	length(LexVars, NVars),
	(
	  NVars =:= 1
	->
	  LexVars = [LexComponent]
	;
	  LexComponent = LexCompnt
	),
	lexico_term(Name, LexVarOrderCriteria, LexComponents).

% order w.r.t. list of criteria 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
order_crit(_, _, [], []) :-
	!.
order_crit(N, OrderList, X, X) :-
	length(OrderList, M),
	N > M,
	!.
order_crit(NCrit, OrderList, [LexValGroup | LexValGroups], OrderedLexVals) :-
				%nl, write('-ncrit-\t\t'), write(NCrit),nl,
	MCrit is NCrit + 1,	
	length(LexValGroup, L),
	nth1(NCrit, OrderList, Order),
				%write('-order-\t\t'), write(Order),nl,
				%write('-lexico-\t'), nl,write(LexValGroup),nl,
	select_criterion(LexValGroup, NCrit, Lex),
				%write('-lex-\t\t'), write(Lex),nl,
	(
	  Order == is
	->
	  keyclumps_is(Lex, OrderedLexClumped)
	;
	  samsort(Order, Lex, OrderedLex),
	  keyclumps(OrderedLex, OrderedLexClumped)
	),
				%write('-keyclumps-\t'),write(OrderedLexClumped),nl,
	order_full_lex(OrderedLexClumped, LexValGroup, OrderedFullLex),
				%write('-orderedFullLex-'), write(OrderedFullLex),nl,nl,
	length(OrderedFullLex, LClumps),
	(
	  LClumps == L
	->
	  OrderedLexValGroup = OrderedFullLex
	;
	  order_crit(MCrit, OrderList, OrderedFullLex, OrderedLexValGroup)
	),
	order_crit(NCrit, OrderList, LexValGroups, OrderedLexValGroups),
	append(OrderedLexValGroup, OrderedLexValGroups, OrderedLexVals).

select_criterion([], _, []).
select_criterion([List-Tag|T], N, [Crit-Tag|Crits]) :-
	nth1(N, List, Crit),
	select_criterion(T, N, Crits).

order_full_lex([], _, []).
order_full_lex([Group|Groups], LexValGroup, [OrderedFullGroup|OrderedFullGroups]) :-
	order_full_lex_group(Group, LexValGroup, OrderedFullGroup),
	order_full_lex(Groups, LexValGroup, OrderedFullGroups).

order_full_lex_group([], _, []).
order_full_lex_group([_-T|Comps], LexValGroup, [C-T|FullComps]) :-
	order_full_lex_group(Comps, LexValGroup, FullComps),
	memberchk(C-T, LexValGroup).

% extract variables from components
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
filter_variables([], []).
filter_variables([H|T], Variables) :-
	is_list(H),
	!,
	filter_variables(H, VHs),
	filter_variables(T, VTs),
	append(VHs, VTs, Variables).
filter_variables([_-(_-V)|T], Variables) :-
	filter_variables(T, Vs),
	append([V], Vs, Variables).

% extract labels-constraints from components
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
filter_constraints([], []) :-
	!.
filter_constraints([H | T], Constraints) :-
	is_list(H),
	!,
	filter_constraints(H, CHs),
	filter_constraints(T, CTs),
	append(CHs, CTs, Constraints).
filter_constraints([_-(_-node(_, H)) | T], Constraints) :-
	!,
	filter_constraints(H, C),
	filter_constraints(T, Cs),
	append(C, Cs, Constraints).
filter_constraints([_-(_-H) | T], Constraints) :-
	!,
	filter_constraints(H, C),
	filter_constraints(T, Cs),
	append(C, Cs, Constraints).
filter_constraints(_-and(H), [and(C)]) :-
	!,
	filter_constraints(H, C).
filter_constraints(_-(_-node(L, C)), [node(L, C)]) :-
	!.
filter_constraints(_-(_-C), [C]) :-
	!.
filter_constraints(C, [C]).

% value ordering
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
order_values([], _, []).
order_values([Var|Vars], ValOrderCiteria, [VarsValLabeling | VarsValLabelings]) :-
	apply_value_choice(Var, ValOrderCiteria, VarsValLabeling),
	order_values(Vars, ValOrderCiteria, VarsValLabelings).

apply_value_choice(Var, [], labeling([], [Var])).
apply_value_choice(free_var(Name, N, I), [ValOrderCrit | ValOrderCrits], VVLabeling) :- 
	subst(^, Name, ValOrderCrit, TermSubst),
	TermSubst =.. [ValChoiceOption, TermToParse],
	parse_expr(TermToParse, TermToRewrite),
				% write(free_var(Name, N, I)),nl,
	on_exception(_Error,
		     (
		       env_new(Env),
		       rewrite_term(TermToRewrite, [], Env, RewrittenTerm),
		       (
			 RewrittenTerm == free_var(Name, N, I)
		       ->			 
			 VVLabeling = labeling([ValChoiceOption],
					       [RewrittenTerm])
		       ;
			 raise_exception(generic_rewriting(not_reduced_to_term(RewrittenTerm),
							   [RewrittenTerm],[]))
		       )
		     ),
		     (
		       apply_value_choice(free_var(Name, N, I), ValOrderCrits, VVLabeling)
		     )
		    ).


merge_labelings([], 1) :-
   !.

merge_labelings([labeling([], LabelingVars) | RestOfLabelings], MergedLabelings) :-	
   !,
   write(merge_labelings(RestOfLabelings)),nl,
   merge_labelings(RestOfLabelings,
                   up - LabelingVars,
                   MergedLabelings).

merge_labelings([labeling([LabelingOption], LabelingVars) | RestOfLabelings], MergedLabelings) :-
   merge_labelings(RestOfLabelings,
                   LabelingOption - LabelingVars,
                   MergedLabelings).

merge_labelings([],
		LabelingOption - LabelingVars,
		(labeling([LabelingOption], LabelingVars))) :-
   !.
merge_labelings([labeling([], LabelingVars) | Labelings],
		up - LabelingVarsSoFar,
		MergedLabelings) :-
   !,
   append(LabelingVarsSoFar, LabelingVars, MergedLabelingVars),	
   merge_labelings(Labelings, up - MergedLabelingVars, MergedLabelings).
merge_labelings([labeling([LabelingOptionSoFar], LabelingVars) | Labelings],
		LabelingOptionSoFar - LabelingVarsSoFar,
		MergedLabelings) :-
   !,
   append(LabelingVarsSoFar, LabelingVars, MergedLabelingVars),	
   merge_labelings(Labelings, LabelingOptionSoFar-MergedLabelingVars, MergedLabelings).
merge_labelings([labeling([], LabelingVars) | Labelings],
		LabelingOptionSoFar - LabelingVarsSoFar,
		(
                   labeling([LabelingOptionSoFar], LabelingVarsSoFar) and
                   MergedLabelingsSoFar
                )
               ) :-
   !,
   merge_labelings(Labelings, up - LabelingVars, MergedLabelingsSoFar).
merge_labelings([labeling([NewLabelingOption], LabelingVars) | Labelings],
		LabelingOptionSoFar - LabelingVarsSoFar,
		(  
                   labeling([LabelingOptionSoFar], LabelingVarsSoFar) and
                   MergedLabelingsSoFar
                )
               ) :-
   merge_labelings(Labelings, NewLabelingOption - LabelingVars, MergedLabelingsSoFar).

% merge_labelings(Labelings, MergedLabelings) :-	
% 	Labelings = [labeling([], LabelingVars) | RestOfLabelings],
% 	!,
% 	merge_labelings(RestOfLabelings, up-LabelingVars, MergedLabelings).
% merge_labelings(Labelings, MergedLabelings) :-	
% 	Labelings = [labeling([LabelingOption], LabelingVars)|RestOfLabelings],
% 	merge_labelings(RestOfLabelings, LabelingOption-LabelingVars, MergedLabelings).

% merge_labelings([],
% 		LabelingOption-LabelingVars,
% 		(labeling([LabelingOption], LabelingVars))) :-
% 	!.
% merge_labelings([labeling([], LabelingVars) | Labelings],
% 		up-LabelingVarsSoFar,
% 		MergedLabelings) :-
% 	!,
% 	append(LabelingVarsSoFar, LabelingVars, MergedLabelingVars),	
% 	merge_labelings(Labelings, up-MergedLabelingVars, MergedLabelings).
% merge_labelings([labeling([LabelingOptionSoFar], LabelingVars) | Labelings],
% 		LabelingOptionSoFar-LabelingVarsSoFar,
% 		MergedLabelings) :-
% 	!,
% 	append(LabelingVarsSoFar, LabelingVars, MergedLabelingVars),	
% 	merge_labelings(Labelings, LabelingOptionSoFar-MergedLabelingVars, MergedLabelings).
% merge_labelings([labeling([NewLabelingOption],LabelingVars)|Labelings],
% 		LabelingOptionSoFar-LabelingVarsSoFar,
% 		(labeling([LabelingOptionSoFar],LabelingVarsSoFar),MergedLabelingsSoFar)) :-
% 	merge_labelings(Labelings, NewLabelingOption-LabelingVars, MergedLabelingsSoFar).

% key clumps
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
keyclumps_is(Lex, [OrderedLexYes, OrderedLexNo]) :-
	keyclumps_is_(Lex, OrderedLexYes, OrderedLexNo).

keyclumps_is_([], [], []) :-
	!.
keyclumps_is_([V-(N-V) | T], Y, Nos) :-
	!,
	keyclumps_is_(T, Yess, Nos),
	append([V-(N-V)], Yess, Y).
keyclumps_is_([H | T], Yess, N) :-
	keyclumps_is_(T, Yess, Nos),
	append([H], Nos,  N).	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% term comparison
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% greatest 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
greatest(T1-X, T2-Y) :-
	X @=< Y,
	T1 @>= T2.

% least 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
least(T1-X, T2-Y) :-
	X @=< Y,
	T1 @=< T2.

% any
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
any(_-X, _-Y) :-
	X @=< Y.

% least term (wrt. the standard order)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
least_term(T) :-                % undefined
   T = -100.0.

% greatest term (wrt. the standard order)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
greatest_term(T) :-             % undefined
   functor(T, zzzzz, 100).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% conjunction or disjunction operator
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
conj_or_disj_op(Term, Op, P, Q) :-
	Term =.. [Op, P, Q],
	memberchk(Op, [and, or]).

first_labels_list([], []).
first_labels_list([_-(_-node(LL, _)) | _LexLabels], LL).

rest_of_labels([], _Connective, []).
rest_of_labels([Connective-_ | Labels], Connective, RestOfLabels) :-
	!,
	rest_of_labels(Labels, Connective, RestOfLabels).
rest_of_labels([C-L | Labels], Connective, [C-L | RestOfLabels]) :-
	rest_of_labels(Labels, Connective, RestOfLabels).
