:- module(geosr, [geosr/3]).

/*
 * $Log: geosr.pl,v $
 * Revision 1.13  2008/03/14 08:40:11  matsc
 * more careful planning of sweeps
 *
 * Revision 1.12  2008/03/04 14:34:07  matsc
 * clustering of conjuncts (equivalence sets)
 *
 * Revision 1.11  2008/02/29 09:05:42  matsc
 * pretty-printing
 *
 * Revision 1.10  2008/02/26 17:28:07  matsc
 * reduce amount of witness seeks
 *
 */

/* Prototype implementation of geosr(Objects,Shapes,Rules)

% Important datatypes:

%%%%%%%%%%%%%%

% Outset (set of (k+1)-dimensional points)

Outset --> list of OBox

%%%%%%%%%%%%%%

% O-box (outbox).

OBox --> list of Integer..Integer

%%%%%%%%%%%%%%

% S-expression (set expression), evaluates to an Outset
% S-expressions are compiled; see below.

SExpr --> {} // empty set 
        | {OExpr} // singleton outbox if nonempty, otherwise empty set
        | gez(AExpr) // if (AExpr>=0) then empty set, else infinite set
        | sids(Oid,J,A,=<,AExpr) // {p in Z^{k+1} | O^sid=p[k+1] implies O^pbox(J)^A =< AExpr}
        | sids(Oid,J,A,>=,AExpr) // {p in Z^{k+1} | O^sid=p[k+1] implies O^pbox(J)^A >= AExpr}
	| SExpr/\SExpr // intersection 
	| SExpr\/SExpr // union 

% O-expression (outbox expression), evaluates to an outbox.

OExpr --> list of AExpr..AExpr

%%%%%%%%%%%%%%

% A-expression (arithmetic expression), evaluates to an integer.

AExpr --> inf | sup | int(Integer)
	| xmin(OID,J)   // min(object(I,_,_,_)^x(J))
	| xmax(OID,J)   // max(object(I,_,_,_)^x(J))
	| pmin(OID,J,A) // min(object(I,_,_,_)^pbox(J)^A)
	| pmax(OID,J,A) // max(object(I,_,_,_)^pbox(J)^A)
	| AExpr + AExpr
	| AExpr - AExpr
	| mul(AExpr,Integer)
	| floordiv(AExpr,Integer)
	| ceildiv(AExpr,Integer)

%%%%%%%%%%%%%%

% Instruction set, i.e. compiled representation of S-expression.
% Register-based virtual machine, for common subexpression elimination.
% With respect to a current object O.
% A register holds an integer or an Outset.
% The end result is the Outset used for sweeping O in the various dimensions.

Procedure --> procedure(NTemp,Dest,Code,Dep) // stores an Outset in register Dest
	                                     // using up to NTemp registers
	                                     // Dep are the objects on which Code depends

Code --> list of InsnPair

InsnPair -->  Insn-Dest           // executes Insn, stores value in register Dest

Insn --> inf | sup | int(I)       // store the relevant integer
       | xmin(OID,J)              // stores min(object(I,_,_,_)^x(J))
       | xmax(OID,J)              // stores max(object(I,_,_,_)^x(J))
       | pmin(OID,J,A)            // stores min(object(I,_,_,_)^pbox(J)^A)
       | pmax(OID,J,A)            // stores max(object(I,_,_,_)^pbox(J)^A)
       | plus(R1,R2)              // stores R1 + R2
       | minus(R1,R2)             // stores R1 - R2
       | mul(R1,I)                // stores R1 * I
       | floordiv(R1,I)           // stores floor(R1 / I)
       | ceiliv(R1,I)             // stores ceiling(R1 / I)
       | obox(list of RegPair)    // stores an Outset of one OBox if nonempty, otherwise the empty set
       | gez(R1)                  // if R1>=0, stores the empty set, otherwise stores dom(O.x)
       | sids(_,J,A,Op,R1)        // stores outset corresponding to the values of O.sid
				  //  that satisfy O^pbox(J)^A Op R1
       | intersection(R1,R2,Code) // if R1 is the empty set, stores the empty set
				  //  otherwise, executes Code with value in R2,
				  //  and stores the intersection of R1 and R2
       | union(R1,R2,Code)        // if R1 includes dom(O.x), stores R1,
				  //  otherwise, executes Code with value in R2,
				  //  and stores the union of R1 and R2

RegPair --> Reg1..Reg2            // builds an interval, part of an OBox

*/

:- use_module(library(avl), [
			avl_change/5,
			avl_del_min/4,
			avl_delete/4,
			avl_fetch/3,
			avl_store/4,
			avl_to_list/2,
			empty_avl/1,
			ord_list_to_avl/2]).

:- use_module(library(lists), [
			keyclumped/2,
			keys_and_values/3,
			max_member/2,
			min_member/2,
			nth0/3,
			nth1/3,
			transpose/2,
			select/3]).

:- use_module(library(ordsets), [
			ord_subtract/3]).

:- use_module(library(trees), [
			get_label/3,
			list_to_tree/2]).

:- use_module(library(types), [
			must_be/4]).

:- use_module(library(clpfd)).

:- op(1200, xfx, --->).

:- multifile     clpfd:dispatch_global/4.
:- discontiguous clpfd:dispatch_global/4.
clpfd:dispatch_global(geosr(_,_,_), f(Obj2Procs1,Src2Targets,K,Objects,Sboxes), f(Obj2Procs2,Src2Targets,K,Objects,Sboxes), Actions) :- !,
        geosr_solver(Objects, Sboxes, Obj2Procs1, Obj2Procs2, Src2Targets, K, Actions, []).

geosr(Objects1, Sboxes1, Rules1) :-
	Goal = geosr(Objects1,Sboxes1,Rules1),
	geosr_sboxes(Sboxes1, Sboxes2, Goal, K, Sids1),
	sort(Sboxes2, Sboxes3),	% by ascending SID
	sort(Sids1, Sids2),
	list_to_fdset(Sids2, SidSet),
	geosr_objects(Objects1, Objects2, SidSet, Goal, K, Susp, []),
	sort(Objects2, Objects3), % by ascending OID
	rewrite_rules(Rules1, Tree1, Objects3, Sboxes3, Goal),
	rewrite_away_not(Tree1, Tree2),
	rewrite_relations(Tree2, Tree3),
	rewrite_away_ops(Tree3, Tree4),
	rewrite_away_min_max(Tree4, Tree5),
	rewrite_linearize(Tree5, Tree6),
	rewrite_simplify(Tree6, Tree7, context(Objects3,Sboxes3)),
	compile_linearized(Tree7, K, ObjProcs1, []),
	total_codelength(ObjProcs1, 0, TotalInsns, 0, TotalIxs),
	print_message(informational, format('TOTAL ~d #indexicals, ~d #insns',[TotalIxs,TotalInsns])),
	keysort(ObjProcs1, ObjProcs2),
	keyclumped(ObjProcs2, ObjProcs3),
	pair_values_with_nil(ObjProcs3, ObjProcs4),
	ord_list_to_avl(ObjProcs4, Obj2Procs),
	all_dependencies(ObjProcs1, SrcTargets1, []),
	sort(SrcTargets1, SrcTargets2),
	keyclumped(SrcTargets2, SrcTargets3),
	ord_list_to_avl(SrcTargets3, Src2Targets),
	fd_global(Goal, f(Obj2Procs,Src2Targets,K,Objects3,Sboxes3), Susp).

pair_values_with_nil([], []).
pair_values_with_nil([K-V|L1], [K-(V-[])|L2]) :-
	pair_values_with_nil(L1, L2).

geosr_objects([], [], _, _, _) --> [].
geosr_objects([object(Oid,Sid,Origin)|Os1], Os2, SidSet, Goal, K) --> !,
	geosr_objects([object(Oid,Sid,Origin,[])|Os1], Os2, SidSet, Goal, K).
geosr_objects([Object|Os1], [Object|Os2], SidSet, Goal, K) -->
	{Object = object(Oid,Sid,Origin1,Atts)},
	{must_be(Oid, integer, Goal, 1)},
	{Sid in_set SidSet},
	{must_be(Atts, proper_list, Goal, 1)},
	{length(Origin1, K)},
	clpfd:finite_dvar_list_susp([Sid],   _, dom,    Goal, 1),
	clpfd:finite_dvar_list_susp(Origin1, _, minmax, Goal, 1),
	geosr_objects(Os1, Os2, SidSet, Goal, K).

geosr_sboxes([], [], _, _, []).
geosr_sboxes([sbox(Sid,Trans,Size)|Ss1], Ss2, Goal, K, Sids) :- !,
	geosr_sboxes([sbox(Sid,Trans,Size,[])|Ss1], Ss2, Goal, K, Sids).
geosr_sboxes([Sbox|Ss1], [Sbox|Ss2], Goal, K, [Sid|Sids]) :-
	Sbox = sbox(Sid,Translation,Size,Atts),
	must_be(Sid, integer, Goal, 2),
	must_be(Translation, list(integer), Goal, 2),
	length(Translation, K),
	must_be(Size, list(integer(>(0))), Goal, 2),
	must_be(Atts, proper_list, Goal, 2),
	length(Size, K),
	geosr_sboxes(Ss1, Ss2, Goal, K, Sids).

geosr_solver(Objects, Sboxes, Obj2Procs1, Obj2Procs2, Src2Targets, K) -->
	{avl_to_list(Obj2Procs1, Queue1)},
	{keys_and_values(Queue1, Queue2, _)},
	{geosr_domains(Objects, Doms1)},
	{geosr_fixpoint(Queue2, Obj2Procs1, Obj2Procs2, Doms1, Doms2, Src2Targets, K, Sboxes)}, !,
	geosr_actions(Doms2).
geosr_solver(_, _, _, _, _, _) --> [fail].

geosr_domains(Objects, Doms) :-
	empty_avl(A),
	geosr_domains(Objects, A, Doms).

geosr_domains([], Doms, Doms).
geosr_domains([object(OID,SID,Xs,_)|Os], Doms0, Doms) :-
	append(Xs, [SID], Xs1),
	geosr_domain(Xs1, 0, OID, Doms0, Doms1),
	geosr_domains(Os, Doms1, Doms).

geosr_domain([], _, _, Doms, Doms).
geosr_domain([X|Xs], I, OID, Doms0, Doms) :-
	fd_set(X, S),
	J is I+1,
	avl_store(x(OID,J), Doms0, X-S, Doms1),
	geosr_domain(Xs, J, OID, Doms1, Doms).

geosr_actions(Doms) -->
	{empty_avl(Doms)}, !.
geosr_actions(Doms1) -->
	{avl_del_min(Doms1, _, X-S2, Doms2)},
	{fd_set(X, S1)},
	({S1==S2} -> [] ; [X in_set S2]),
	geosr_actions(Doms2).

geosr_fixpoint([],           Obj2Procs,  Obj2Procs,  Doms,  Doms,  _,           _, _).
geosr_fixpoint([Obj|Queue1], Obj2Procs1, Obj2Procs4, Doms1, Doms3, Src2Targets, K, Sboxes) :-
	avl_change(Obj, Obj2Procs1, SExprs1-Wits1, Obj2Procs2, SExprs2-Wits2), !,
	sweep_object(Obj, SExprs1, Wits1, SExprs2, Wits2, Doms1, Doms2, Cmp, CurDom, K, Sboxes),
	(   Cmp==(=) -> Queue3 = Queue1
	;   avl_fetch(Obj, Src2Targets, Targets1) ->
	    sort(Queue1, Queue2),
	    ord_subtract(Targets1, Queue2, Targets2),
	    append(Queue1, Targets2, Queue3)
	;   Queue3 = Queue1
	),
	(   ground_cur_dom(CurDom) ->
	    avl_delete(Obj, Obj2Procs2, _, Obj2Procs3)
	;   Obj2Procs3 = Obj2Procs2
	),
	geosr_fixpoint(Queue3, Obj2Procs3, Obj2Procs4, Doms2, Doms3, Src2Targets, K, Sboxes).
geosr_fixpoint([_|Queue1], Obj2Procs1, Obj2Procs4, Doms1, Doms3, Src2Targets, K, Sboxes) :-
	geosr_fixpoint(Queue1, Obj2Procs1, Obj2Procs4, Doms1, Doms3, Src2Targets, K, Sboxes).

sweep_object(o(OID), SExprs1, Wits1, SExprs2, Wits3, Doms1, Doms2, Cmp, CurDom3, K, Sboxes) :-
	print_message(informational, format('sweeping object ~d',[OID])),
	K1 is K+1,
	fetch_cur_dom(0, K1, OID, Doms1, Xs, CurDom1),
	eval_sexprs(SExprs1, SExprs2, CurDom1, CurDom2, Doms1, Forbidden, Sboxes),
	filter_witnesses(Wits1, CurDom2, Forbidden, Wits2),
	plan_sweep(CurDom2, 0, K, 0, Plan, []),
	do_sweep(Plan, CurDom2, CurDom3, Forbidden, Wits2, Wits3),
	store_cur_dom(0, K1, OID, Doms1, Xs, CurDom3, Doms2),
	compare(Cmp, CurDom1, CurDom3).

plan_sweep([], _, _, _) --> [].
plan_sweep([Dom|Doms], I, K, Flag) -->
	{fdset_singleton(_, Dom)},
	{I<K ; Flag>0}, !,
	{J is I+1},
	plan_sweep(Doms, J, K, Flag).
plan_sweep([_|Doms], I, K, _) -->
	{J is I+1},
	plan_item(J, K),
	plan_sweep(Doms, J, K, 1).

plan_item(J, K) --> ({J>K} -> [J-domain] ; [J-bound]).

do_sweep([], CurDom, CurDom, _, Wits, Wits).
do_sweep([J-domain|Plan], CurDom0, CurDom, Forbidden, Wits0, Wits) :- !,
	K is J-1,
	nth0(K, CurDom0, SidSet1),
	fdset_to_list(SidSet1, Sids1),
	sweep_domain(Sids1, Sids2, CurDom0, Forbidden, K, Wits0, Wits1),
	(   Sids2==[] -> fail
	;   Sids2==Sids1 -> CurDom1 = CurDom0
	;   list_to_fdset(Sids2, SidSet2),
	    replast(CurDom0, SidSet2, CurDom1)
	),
	do_sweep(Plan, CurDom1, CurDom, Forbidden, Wits1, Wits).
do_sweep([J-bound|Plan], CurDom0, CurDom, Forbidden, Wits0, Wits) :-
	I is J-1,
	sweep_up(I, CurDom0, CurDom1, Forbidden, Wits0, Wits1),
	sweep_down(I, CurDom1, CurDom2, Forbidden, Wits1, Wits2),
	do_sweep(Plan, CurDom2, CurDom, Forbidden, Wits2, Wits).


sweep_domain([], [], _, _, _, Wits, Wits).
sweep_domain([S|Sids1], [S|Sids2], CurDom1, Forbidden, K, Wits0, Wits) :-
	value_has_support(S, K, Wits0), !,
	sweep_domain(Sids1, Sids2, CurDom1, Forbidden, K, Wits0, Wits).
sweep_domain([S|Sids1], [S|Sids2], CurDom1, Forbidden, K, Wits0, Wits) :-
	fdset_singleton(S1, S),
	replast(CurDom1, S1, CurDom2),
	sweep_up_init(CurDom2, First, Next),
	do_sweep_up(First, Next, void, Witness, CurDom2, K, Forbidden), !,
	(Witness==void -> Wits1 = Wits0 ; Wits1 = [Witness|Wits0]),
	sweep_domain(Sids1, Sids2, CurDom1, Forbidden, K, Wits1, Wits).
sweep_domain([_|Sids1], Sids2, CurDom1, Forbidden, K, Wits0, Wits) :-
	sweep_domain(Sids1, Sids2, CurDom1, Forbidden, K, Wits0, Wits).

filter_witnesses([], _, _, []).
filter_witnesses([W|L1], CurDom, Forbidden, L2) :-
	\+point_in_domain(W, CurDom), !,
	filter_witnesses(L1, CurDom, Forbidden, L2).
filter_witnesses([W|L1], CurDom, Forbidden, L2) :-
	member(OBox, Forbidden),
	point_in_obox(W, OBox), !,
	filter_witnesses(L1, CurDom, Forbidden, L2).
filter_witnesses([W|L1], CurDom, Forbidden, [W|L2]) :-
	filter_witnesses(L1, CurDom, Forbidden, L2).

point_in_domain([], []).
point_in_domain([X|Xs], [S|Ss]) :-
	fdset_member(X, S),
	point_in_domain(Xs, Ss).

value_has_support(Val, I, Wits) :-
	Wits = [W|_],
	wit_template(W, W1, 0, I, Val),
	memberchk(W1, Wits).

wit_template([],     [],     _, _, _).
wit_template([_|W1], [V|W2], I, I, V) :- !,
	J is I+1,
	wit_template(W1, W2, J, I, V).
wit_template([_|W1], [_|W2], I, K, V) :-
	J is I+1,
	wit_template(W1, W2, J, K, V).

sweep_up(I, CurDom1, CurDom2, Forbidden, Wits0, Wits) :-
	wit_init_up(Wits0, W0, I),
	sweep_up_init(CurDom1, First, Next),
	do_sweep_up(First, Next, W0, Witness, CurDom1, I, Forbidden),
	(   Witness==void ->
	    W1 = W0,
	    Wits = Wits0
	;   Wits = [Witness|Wits0],
	    nth0(I, Witness, W1)
	),
	fdset_interval(Mask, W1, sup),
	nth0(I, CurDom1, DomI1),
	fdset_intersection(DomI1, Mask, DomI2),
	setnth0(I, CurDom1, DomI2, CurDom2).

wit_init_up([], void, _).
wit_init_up(Wits0, W, I) :- !,
	transpose(Wits0, WitsR),
	nth0(I, WitsR, Col),
	min_member(W, Col).

sweep_up_init([], [], []).
sweep_up_init([S|Ss], [F|Fs], [N|Ns]) :-
	fdset_min(S, F),
	fdset_max(S, N1),
	N is N1+1,
	sweep_up_init(Ss, Fs, Ns).

do_sweep_up(Cand0, _,     W, void   , _     , I, _        ) :-
	W\==void,
	nth0(I, Cand0, W), !.
do_sweep_up(Cand0, Next0, W, Witness, CurDom, I, Forbidden) :-
	member(OBox, Forbidden),
	point_in_obox(Cand0, OBox), !,
	point_after_obox(Next0, OBox, Next1),
	length(Cand0, N),
	next_cand_up(Cand0, Cand1, Next1, Next, CurDom, I, N, N),
	do_sweep_up(Cand1, Next, W, Witness, CurDom, I, Forbidden).
do_sweep_up(Witness, _, _, Witness, _, _, _) :-
	print_message(informational, format('found witness ~w',[Witness])),
	true.

next_cand_up(Cand0, Cand, Next0, Next, CurDom, D, JJ, K) :-
	JJ>0,
	J is JJ-1,
	J1 is (J+D) mod K,
	nth0(J1, CurDom, DomJ1),
	fdset_min(DomJ1, MinJ1),
	fdset_max(DomJ1, MaxJ1),
	nth0(J1, Next0, NJ1),
	MaxJ11 is MaxJ1+1,
	setnth0(J1, Next0, MaxJ11, Next1),
	(   NJ1=<MaxJ1 ->
	    setnth0(J1, Cand0, NJ1, Cand), 
	    Next = Next1
	;   setnth0(J1, Cand0, MinJ1, Cand1),
	    next_cand_up(Cand1, Cand, Next1, Next, CurDom, D, J, K)
	).

sweep_down(I, CurDom1, CurDom2, Forbidden, Wits0, Wits) :-
	wit_init_down(Wits0, W0, I),
	sweep_down_init(CurDom1, First, Next),
	do_sweep_down(First, Next, W0, Witness, CurDom1, I, Forbidden),
	(   Witness==void ->
	    W1 = W0,
	    Wits = Wits0
	;   Wits = [Witness|Wits0],
	    nth0(I, Witness, W1)
	),
	fdset_interval(Mask, inf, W1),
	nth0(I, CurDom1, DomI1),
	fdset_intersection(DomI1, Mask, DomI2),
	setnth0(I, CurDom1, DomI2, CurDom2).

wit_init_down([], void, _).
wit_init_down(Wits0, W, I) :- !,
	transpose(Wits0, WitsR),
	nth0(I, WitsR, Col),
	max_member(W, Col).

sweep_down_init([], [], []).
sweep_down_init([S|Ss], [F|Fs], [N|Ns]) :-
	fdset_max(S, F),
	fdset_min(S, N1),
	N is N1-1,
	sweep_down_init(Ss, Fs, Ns).

do_sweep_down(Cand0, _,     W, void   , _     , I, _        ) :-
	W\==void,
	nth0(I, Cand0, W), !.
do_sweep_down(Cand0, Next0, W, Witness, CurDom, I, Forbidden) :-
	member(OBox, Forbidden),
	point_in_obox(Cand0, OBox), !,
	point_before_obox(Next0, OBox, Next1),
	length(Cand0, N),
	next_cand_down(Cand0, Cand1, Next1, Next, CurDom, I, N, N),
	do_sweep_down(Cand1, Next, W, Witness, CurDom, I, Forbidden).
do_sweep_down(Witness, _, _, Witness, _, _, _) :-
	print_message(informational, format('found witness ~w',[Witness])),
	true.

next_cand_down(Cand0, Cand, Next0, Next, CurDom, D, JJ, K) :-
	JJ>0,
	J is JJ-1,
	J1 is (J+D) mod K,
	nth0(J1, CurDom, DomJ1),
	fdset_min(DomJ1, MinJ1),
	fdset_max(DomJ1, MaxJ1),
	nth0(J1, Next0, NJ1),
	MinJ11 is MinJ1-1,
	setnth0(J1, Next0, MinJ11, Next1),
	(   NJ1>=MinJ1 ->
	    setnth0(J1, Cand0, NJ1, Cand), 
	    Next = Next1
	;   setnth0(J1, Cand0, MaxJ1, Cand1),
	    next_cand_down(Cand1, Cand, Next1, Next, CurDom, D, J, K)
	).
	
obox_intersection([], [], []).
obox_intersection([A1..B1|OE1], [A2..B2|OE2], [A3..B3|OE3]) :-
	A3 is max(A1,A2),
	B3 is min(B1,B2),
	A3=<B3,
	obox_intersection(OE1, OE2, OE3).

obox_subset([], []).
obox_subset([A1..B1|OE1], [A2..B2|OE2]) :-
	A1>=A2,
	B1=<B2,
	obox_subset(OE1, OE2).

% SOS ~= forbidden set used in the sweep
% A SExpr can be forgotten once its RHS is ground and the SOS it generates
% is disjoint from the origin domain box.
eval_sexprs([], [], CurDom, CurDom, _, [], _).
eval_sexprs([SExpr|SExprs1], SExprs3, CurDom1, CurDom3, Doms, SOS4, Sboxes) :- 
	dom_bbox(CurDom1, BBox),
	eval_sexpr(SExpr, BBox, Doms, SOS1, SeenVar, Sboxes),
	prune_from_sos(SOS1, SOS2, CurDom1, CurDom2),
	eval_sexprs(SExprs1, SExprs2, CurDom2, CurDom3, Doms, SOS3, Sboxes),
	(   var(SeenVar), SOS2==[] -> SExprs3 = SExprs2
	;   SExprs3 = [SExpr|SExprs2]
	),
	sos_union(SOS2, SOS3, SOS4, [BBox]).

prune_from_sos([], [], CurDom, CurDom).
prune_from_sos([OBox|SOS1], SOS2, CurDom1, CurDom3) :-
	do_prune_from_sos(OBox, CurDom1, CurDom2, 0), !,
	prune_from_sos(SOS1, SOS2, CurDom2, CurDom3).
prune_from_sos([OBox|SOS1], [OBox|SOS2], CurDom1, CurDom2) :-
	prune_from_sos(SOS1, SOS2, CurDom1, CurDom2).

% Pre-sweep pruning.
% Suppose 1st arg [forbidden region] is (1..10)x(1..10)x(5..5),
%         2nd arg [origin domain] is (3..7)x(3..7)x(1..10).
% Then 5 can be deleted from x(3) and this forbidden region can be forgotten
do_prune_from_sos([], [], [], 1).
do_prune_from_sos([O1..O2|OBox], [D1|Dom1], [D2|Dom2], C1) :-
	fdset_interval(O, O1, O2),
	(   fdset_subset(D1, O) -> D2 = D1, C2 = C1
	;   C1=:=0,
	    C2 = 1,
	    fdset_subtract(D1, O, D2)
	),
	do_prune_from_sos(OBox, Dom1, Dom2, C2).

eval_sexpr(procedure(NTemp,Dest,Insns,_), BBox, Doms, SOS, SeenVar, Sboxes) :-
	allocate(NTemp, Regs),
	get_register(Regs, Dest, SOS),
	exec_insns(Insns, context(BBox,Doms,SeenVar,Regs,Sboxes)).

dom_bbox([], []).
dom_bbox([Dom|Doms], [Min..Max|BBox]) :-
	fdset_min(Dom, Min),
	fdset_max(Dom, Max),
	dom_bbox(Doms, BBox).

exec_insns([], _).
exec_insns([Insn-Dest|Insns], Context) :-
	exec_insn(Insn, Dest, Insns, Context).

exec_insn(inf, Dest, Insns, Context) :-
	Context = context(_,_,_,Regs,_),
	get_register(Regs, Dest, -0x10000000),
	exec_insns(Insns, Context).
exec_insn(sup, Dest, Insns, Context) :-
	Context = context(_,_,_,Regs,_),
	get_register(Regs, Dest,  0x10000000),
	exec_insns(Insns, Context).
exec_insn(int(I), Dest, Insns, Context) :-
	Context = context(_,_,_,Regs,_),
	get_register(Regs, Dest, I),
	exec_insns(Insns, Context).
exec_insn(xmin(I,J), Dest, Insns, Context) :-
	Context = context(_,Doms,SeenVar,Regs,_),
	avl_fetch(x(I,J), Doms, _-S),
	(fdset_singleton(S,_) -> true ; SeenVar=true),
	fdset_min(S, Val),
	get_register(Regs, Dest, Val),
	exec_insns(Insns, Context).
exec_insn(xmax(I,J), Dest, Insns, Context) :-
	Context = context(_,Doms,SeenVar,Regs,_),
	avl_fetch(x(I,J), Doms, _-S),
	(fdset_singleton(S,_) -> true ; SeenVar=true),
	fdset_max(S, Val),
	get_register(Regs, Dest, Val),
	exec_insns(Insns, Context).
exec_insn(pmin(I,J,A), Dest, Insns, Context) :-
	Context = context(BBox,Doms,SeenVar,Regs,Sboxes),
	length(BBox, K1),
	avl_fetch(x(I,K1), Doms, _-S),
	(fdset_singleton(S,_) -> true ; SeenVar=true),
	fdset_to_list(S, L),
	select_jth_sbox_atts(L, J, A, Sboxes, Values),
	min_member(Val, Values),
	get_register(Regs, Dest, Val),
	exec_insns(Insns, Context).
exec_insn(pmax(I,J,A), Dest, Insns, Context) :-
	Context = context(BBox,Doms,SeenVar,Regs,Sboxes),
	length(BBox, K1),
	avl_fetch(x(I,K1), Doms, _-S),
	(fdset_singleton(S,_) -> true ; SeenVar=true),
	fdset_to_list(S, L),
	select_jth_sbox_atts(L, J, A, Sboxes, Values),
	max_member(Val, Values),
	get_register(Regs, Dest, Val),
	exec_insns(Insns, Context).
exec_insn(obox(Args), Dest, Insns, Context) :-
	Context = context(BBox,_,_,Regs,_),
	(make_obox(Args, BBox, OBox, Regs) -> Val = [OBox] ; Val = []),
	get_register(Regs, Dest, Val),
	exec_insns(Insns, Context).
exec_insn(intersection(X,Y,Sub), Dest, Insns, Context) :-
	Context = context(BBox,_,_,Regs,_),
	get_register(Regs, X, X1),
	(   X1==[] -> Val = []
	;   exec_insns(Sub, Context),
	    get_register(Regs, Y, Y1),
	    sos_intersection(X1, Y1, Val, [BBox])
	),
	get_register(Regs, Dest, Val),
	exec_insns(Insns, Context).
exec_insn(union(X,Y,Sub), Dest, Insns, Context) :-
	Context = context(BBox,_,_,Regs,_),
	Top = [BBox],
	get_register(Regs, X, X1),
	(   X1==Top -> Val = Top
	;   exec_insns(Sub, Context),
	    get_register(Regs, Y, Y1),
	    sos_union(X1, Y1, Val, Top)
	),
	get_register(Regs, Dest, Val),
	exec_insns(Insns, Context).
exec_insn(plus(X,Y), Dest, Insns, Context) :-
	Context = context(_,_,_,Regs,_),
	get_register(Regs, X, X1),
	get_register(Regs, Y, Y1),
	Val is X1+Y1,
	get_register(Regs, Dest, Val),
	exec_insns(Insns, Context).
exec_insn(minus(X,Y), Dest, Insns, Context) :-
	Context = context(_,_,_,Regs,_),
	get_register(Regs, X, X1),
	get_register(Regs, Y, Y1),
	Val is X1-Y1,
	get_register(Regs, Dest, Val),
	exec_insns(Insns, Context).
exec_insn(mul(X,Int), Dest, Insns, Context) :-
	Context = context(_,_,_,Regs,_),
	get_register(Regs, X, X1),
	Val is X1*Int,
	get_register(Regs, Dest, Val),
	exec_insns(Insns, Context).
exec_insn(ceildiv(X,Int), Dest, Insns, Context) :-
	Context = context(_,_,_,Regs,_),
	get_register(Regs, X, X1),
	Val is ceiling(X1/Int),
	get_register(Regs, Dest, Val),
	exec_insns(Insns, Context).
exec_insn(floordiv(X,Int), Dest, Insns, Context) :-
	Context = context(_,_,_,Regs,_),
	get_register(Regs, X, X1),
	Val is floor(X1/Int),
	get_register(Regs, Dest, Val),
	exec_insns(Insns, Context).
exec_insn(gez(X), Dest, Insns, Context) :-
	Context = context(BBox,_,_,Regs,_),
	get_register(Regs, X, X1),
	(X1 >= 0 -> Val = [] ; Val = [BBox]),
	get_register(Regs, Dest, Val),
	exec_insns(Insns, Context).
exec_insn(sids(Oid,J,A,Op,X), Dest, Insns, Context) :-
	Context = context(BBox,Doms,_,Regs,Sboxes),
	get_register(Regs, X, X1),
	length(BBox, K1),
	avl_fetch(x(Oid,K1), Doms, _-Set1),
	fdset_to_list(Set1, L1),
	select_jth_sbox_atts(L1, J, A, Sboxes, Values),
	filter_sids(L1, Values, L2, Op, X1),
	list_to_fdset(L2, Set2),
	sids_outset(Set2, BBox, Val, []),
	get_register(Regs, Dest, Val),
	exec_insns(Insns, Context).

filter_sids([], [], [], _, _).
filter_sids([S|L1], [V|Vs], [S|L2], Op, B) :-
	(Op==(=<), V=<B ; Op==(>=), V>=B), !,
	filter_sids(L1, Vs, L2, Op, B).
filter_sids([_|L1], [_|Vs], L2, Op, B) :-
	filter_sids(L1, Vs, L2, Op, B).

sids_outset([], _) --> [].
sids_outset([[Min|Max]|Set], BBox1) --> [BBox2],
	{replast(BBox1, Min..Max, BBox2)},
	sids_outset(Set, BBox1).

replast([_], Y, [Y]) :- !.
replast([X|Xs], Y, [X|Zs]) :-
	replast(Xs, Y, Zs).
	
make_obox([],        [],          [],          _).
make_obox([A..B|L1], [A2..B2|L2], [A3..B3|L3], Regs) :-
	get_register(Regs, A, A1),
	get_register(Regs, B, B1),
	A3 is max(A1,A2),
	B3 is min(B1,B2),
	A3=<B3,
	make_obox(L1, L2, L3, Regs).

allocate(NTemp, functor(Regs)) :-
	NTemp < 256, !,
	functor(Regs, r, NTemp).
allocate(NTemp, bregs(Regs)) :-
	length(Scratch, NTemp),
	list_to_tree(Scratch, Regs).

get_register(functor(Regs), Key, Val) :-
	arg(Key, Regs, Val).
get_register(bregs(Regs), Key, Val) :-
	get_label(Key, Regs, Val).

sos_intersection([], _, [], _) :- !.
sos_intersection(_, [], [], _) :- !.
sos_intersection(Top, SOS, SOS, Top) :- !.
sos_intersection(SOS, Top, SOS, Top) :- !.
sos_intersection([X|Xs], Ys, SOS3, Top) :-
	sos_intersection1(Ys, X, SOS1),
	sos_intersection2(Xs, Ys, SOS2, Top),
	sos_union(SOS1, SOS2, SOS3, Top).

sos_intersection2([], _, [], _).
sos_intersection2([X|Xs], Ys, SOS3, Top) :-
	sos_intersection1(Ys, X, SOS1),
	sos_intersection2(Xs, Ys, SOS2, Top),
	sos_union(SOS1, SOS2, SOS3, Top).

sos_intersection1([], _, []).
sos_intersection1([X|Xs], Y, SOS2) :-
	sos_intersection1(Xs, Y, SOS1),
	(   obox_intersection(X, Y, Z) ->
	    sos_add_element(SOS1, Z, SOS2)
	;   SOS2 = SOS1
	).

sos_union([], SOS, SOS, _) :- !.
sos_union(SOS, [], SOS, _) :- !.
sos_union(Top, _, Top, Top) :- !.
sos_union(_, Top, Top, Top) :- !.
sos_union([X|Xs], SOS1, SOS3, _) :-
	sos_add_element(SOS1, X, SOS2),
	sos_union2(Xs, SOS2, SOS3).
	
sos_union2([], SOS, SOS).
sos_union2([X|Xs], SOS1, SOS3) :-
	sos_add_element(SOS1, X, SOS2),
	sos_union2(Xs, SOS2, SOS3).

sos_add_element([], X, [X]).
sos_add_element([X|SOS1], Y, SOS2) :-
	obox_subset(X, Y), !,
	sos_add_element(SOS1, Y, SOS2).
sos_add_element([X|SOS1], Y, [X|SOS1]) :-
	obox_subset(Y, X), !.
sos_add_element([X|SOS1], Y, [X|SOS2]) :-
	sos_add_element(SOS1, Y, SOS2).

simple_aexpr(inf, inf, -0x10000000).
simple_aexpr(sup, sup,  0x10000000).
simple_aexpr(int(I), I, I).

fetch_cur_dom(N, N, _, _, [], []) :- !.
fetch_cur_dom(I, N, OID, Doms, [X|Xs], [S|CurDom]) :-
	J is I+1,
	avl_fetch(x(OID,J), Doms, X-S),
	fetch_cur_dom(J, N, OID, Doms, Xs, CurDom).

store_cur_dom(N, N, _, Doms, [], [], Doms) :- !.
store_cur_dom(I, N, OID, Doms1, [X|Xs], [S|CurDom], Doms3) :-
	J is I+1,
	avl_store(x(OID,J), Doms1, X-S, Doms2),
	store_cur_dom(J, N, OID, Doms2, Xs, CurDom, Doms3).

ground_cur_dom([]).
ground_cur_dom([S|Ss]) :-
	fdset_singleton(S, _),
	ground_cur_dom(Ss).

point_in_obox([], []).
point_in_obox([X|Xs], [A..B|OBox]) :-
	A=<X, X=<B,
	point_in_obox(Xs, OBox).

point_after_obox([], [], []).
point_after_obox([X|Xs], [_..B|Ys], [Z|Zs]) :-
	Z is min(X,B+1),
	point_after_obox(Xs, Ys, Zs).

point_before_obox([], [], []).
point_before_obox([X|Xs], [A.._|Ys], [Z|Zs]) :-
	Z is max(X,A-1),
	point_before_obox(Xs, Ys, Zs).

setnth0(0, [_|Tail], Elem, [Elem|Tail]) :- !.
setnth0(J, [X|Tail], Elem, [X|Rest]) :-
	I is J-1,
	setnth0(I, Tail, Elem, Rest).

% rewrite_rules(Rules, Tree, Objects, Sboxes, Goal)
%   Expand away all Head--->Body, forall, exists, card, fold, #=>, #<=>, objects(_), sboxes(_)
%   Fold ground _integer_ expressions.
%   Simplify as far as possible.

rewrite_rules(Forms, Expanded, Objects, Sboxes, Goal) :-
	partition_rules(Forms, Rules, Defs),
	length(Rules, N),
	cardify(N, N, N, Rules, Rule),
	rewrite_expr(Rule, Expanded, Defs, Objects, Sboxes, Goal).

partition_rules([], [], []).
partition_rules([X|Xs], Rules, [X|Defs]) :-
	X = (_--->_), !,
	partition_rules(Xs, Rules, Defs).
partition_rules([X|Xs], [X|Rules], Defs) :-
	partition_rules(Xs, Rules, Defs).

rewrite_expr(X, Y, _, _, _, _) :-
	var(X), !, Y = X.
rewrite_expr(X, Y, _, _, _, _) :-
	integer(X), !, Y = X.
rewrite_expr(true, Y, _, _, _, _) :- !,
	Y = true.
rewrite_expr(false, Y, _, _, _, _) :- !,
	Y = false.
rewrite_expr(X, Y, _, _, _, _) :-
	X = object(_,_,_,_), !, Y = X.
rewrite_expr(X, Y, _, _, _, _) :-
	X = sbox(_,_,_,_), !, Y = X.
rewrite_expr(Main^Sel, Part, _, Objects, Sboxes, _) :-
	get_path(Main, Sel, Part, Objects, Sboxes),
	integer(Part), !.
rewrite_expr(Main^Sel, Part, Defs, Objects, Sboxes, Goal) :-
	on_exception(_, rewrite_expr(Sel, Sel1, Defs, Objects, Sboxes, Goal), fail),
	Sel\==Sel1, !,
	rewrite_expr(Main^Sel1, Part, Defs, Objects, Sboxes, Goal).
rewrite_expr(Main^Sel, Main^Sel, _, _, _, _) :- !.
rewrite_expr(#\ P, Expanded, Defs, Objects, Sboxes, Goal) :- !,
	rewrite_expr(P, P1, Defs, Objects, Sboxes, Goal),
	(   P1==false -> Expanded=true
	;   P1==true -> Expanded=false
	;   Expanded=(#\P1)
	).
rewrite_expr(P #/\ Q, Expanded, Defs, Objects, Sboxes, Goal) :- !,
	rewrite_expr(P, P1, Defs, Objects, Sboxes, Goal),
	(   P1==false -> Expanded=false
	;   rewrite_expr(Q, Q1, Defs, Objects, Sboxes, Goal),
	    (   P1==true -> Expanded=Q1
	    ;   Q1==true -> Expanded=P1
	    ;   Q1==false -> Expanded=false
	    ;   Expanded=(P1#/\Q1)
	    )
	).
rewrite_expr(P #\/ Q, Expanded, Defs, Objects, Sboxes, Goal) :- !,
	rewrite_expr(P, P1, Defs, Objects, Sboxes, Goal),
	(   P1==true -> Expanded=true
	;   rewrite_expr(Q, Q1, Defs, Objects, Sboxes, Goal),
	    (   P1==false -> Expanded=Q1
	    ;   Q1==true -> Expanded=true
	    ;   Q1==false -> Expanded=P1
	    ;   Expanded=(P1#\/Q1)
	    )
	).
rewrite_expr(P #=> Q, Expanded, Defs, Objects, Sboxes, Goal) :- !,
	rewrite_expr(#\ P #\/ Q, Expanded, Defs, Objects, Sboxes, Goal).
rewrite_expr(P #<=> Q, Expanded, Defs, Objects, Sboxes, Goal) :- !,
	rewrite_expr(P, P1, Defs, Objects, Sboxes, Goal),
	rewrite_expr(Q, Q1, Defs, Objects, Sboxes, Goal),
	(   P1==true  -> Expanded = Q1
	;   Q1==true  -> Expanded = P1
	;   P1==Q1    -> Expanded = true
	;   P1==false -> Expanded = (#\Q1)
	;   Q1==false -> Expanded = (#\P1)
	;   Expanded = ((P1#/\Q1)#\/((#\P1)#/\(#\Q1)))
	).
rewrite_expr(forall(Quant,List,Expr), Expanded, Defs, Objects, Sboxes, Goal) :- !,
	rewrite_each(List, Quant, Expr, ExpandedList, Defs, Objects, Sboxes, Goal),
	length(ExpandedList, N),
	cardify(N, N, N, ExpandedList, Expanded).
rewrite_expr(exists(Quant,List,Expr), Expanded, Defs, Objects, Sboxes, Goal) :- !,
	rewrite_each(List, Quant, Expr, ExpandedList, Defs, Objects, Sboxes, Goal),
	length(ExpandedList, N),
	cardify(1, N, N, ExpandedList, Expanded).
rewrite_expr(card(Quant,List,L,U,Expr), Expanded, Defs, Objects, Sboxes, Goal) :- !,
	must_be(L, integer, Goal, 0),
	must_be(U, integer, Goal, 0),
	rewrite_each(List, Quant, Expr, ExpandedList, Defs, Objects, Sboxes, Goal),
	length(ExpandedList, N),
	cardify(L, U, N, ExpandedList, Expanded).
rewrite_expr(fold(Quant,List,Op,Ident,Expr), Expanded, Defs, Objects, Sboxes, Goal) :- !,
	rewrite_each(List, Quant, Expr, ExpandedList, Defs, Objects, Sboxes, Goal),
	foldr(ExpandedList, Op, Ident, Expanded).
rewrite_expr(min(X,Y), Expanded, Defs, Objects, Sboxes, Goal) :- !,
	rewrite_expr(X, X1, Defs, Objects, Sboxes, Goal),
	rewrite_expr(Y, Y1, Defs, Objects, Sboxes, Goal),
	(   integer(X1), integer(Y1) -> Expanded is min(X1,Y1)
	;   Expanded = min(X1,Y1)
	).
rewrite_expr(max(X,Y), Expanded, Defs, Objects, Sboxes, Goal) :- !,
	rewrite_expr(X, X1, Defs, Objects, Sboxes, Goal),
	rewrite_expr(Y, Y1, Defs, Objects, Sboxes, Goal),
	(   integer(X1), integer(Y1) -> Expanded is max(X1,Y1)
	;   Expanded = max(X1,Y1)
	).
rewrite_expr(X+Y, Expanded, Defs, Objects, Sboxes, Goal) :- !,
	rewrite_expr(X, X1, Defs, Objects, Sboxes, Goal),
	rewrite_expr(Y, Y1, Defs, Objects, Sboxes, Goal),
	(   X1==0 -> Expanded = Y1
	;   Y1==0 -> Expanded = X1
	;   integer(X1), integer(Y1) -> Expanded is X1+Y1
	;   Expanded = (X1+Y1)
	).
rewrite_expr(X-Y, Expanded, Defs, Objects, Sboxes, Goal) :- !,
	rewrite_expr(X, X1, Defs, Objects, Sboxes, Goal),
	rewrite_expr(Y, Y1, Defs, Objects, Sboxes, Goal),
	(   Y1==0 -> Expanded = X1
	;   integer(X1), integer(Y1) -> Expanded is X1-Y1
	;   Expanded = (X1-Y1)
	).
rewrite_expr(X*Y, Expanded, Defs, Objects, Sboxes, Goal) :- !,
	rewrite_expr(X, X1, Defs, Objects, Sboxes, Goal),
	rewrite_expr(Y, Y1, Defs, Objects, Sboxes, Goal),
	(   X1==0 -> Expanded = 0
	;   Y1==0 -> Expanded = 0
	;   X1==1 -> Expanded = Y1
	;   Y1==1 -> Expanded = X1
	;   integer(X1), integer(Y1) -> Expanded is X1*Y1
	;   Expanded = (X1*Y1)
	).
rewrite_expr(X/Y, Expanded, Defs, Objects, Sboxes, Goal) :- !,
	rewrite_expr(X, X1, Defs, Objects, Sboxes, Goal),
	rewrite_expr(Y, Y1, Defs, Objects, Sboxes, Goal),
	(   Y1==0 -> Expanded = (X1/Y1)
	;   X1==0 -> Expanded = 0
	;   Y1==1 -> Expanded = X1
	;   integer(X1), integer(Y1), X1 mod Y1 =:= 0 ->
	    Expanded is X1//Y1
	;   Expanded = (X1/Y1)
	).
rewrite_expr(Expr, Expanded, Defs, Objects, Sboxes, Goal) :-
	arithop(Expr, Op, X, Y), !,
	rewrite_expr(X, X1, Defs, Objects, Sboxes, Goal),
	rewrite_expr(Y, Y1, Defs, Objects, Sboxes, Goal),
	(   integer(X1), integer(Y1) -> evalop(Op, X1, Y1, Expanded)
	;   Expanded =.. [Op,X1,Y1]
	).
rewrite_expr(Expr, Expanded, Defs, Objects, Sboxes, Goal) :-
	functor(Expr, F, A),
	functor(Head, F, A),
	member((Head--->Body), Defs), !,
	copy_term((Head--->Body), (Expr--->Copy)),
	rewrite_expr(Copy, Expanded, Defs, Objects, Sboxes, Goal).
rewrite_expr(Expr, _Expanded, _Defs, _Objects, _Sboxes, Goal) :-
	raise_exception(existence_error(Goal,3,macro,Expr,0)).

foldr([], _, Z, Z).
foldr([X|Xs], Op, Y, Z) :-
	OpXY =.. [Op,X,Y],
	foldr(Xs, Op, OpXY, Z).

rewrite_args(A, A, _, _, _, _, _, _) :- !.
rewrite_args(I, A, Expr, Expanded, Defs, Objects, Sboxes, Goal) :-
	J is I+1,
	arg(J, Expr, X),
	arg(J, Expanded, Y),
	rewrite_expr(X, Y, Defs, Objects, Sboxes, Goal),
	rewrite_args(J, A,Expr, Expanded, Defs, Objects, Sboxes, Goal).

rewrite_each(objects(Oids), Quant, Expr, ExpandedList, Defs, Objects, Sboxes, Goal) :- !,
	select_objects(Oids, Objects, List),
	rewrite_each1(List, Quant, Expr, ExpandedList, Defs, Objects, Sboxes, Goal).
rewrite_each(sboxes(Sids), Quant, Expr, ExpandedList, Defs, Objects, Sboxes, Goal) :- !,
	select_sboxes(Sids, Sboxes, List, Objects),
	rewrite_each1(List, Quant, Expr, ExpandedList, Defs, Objects, Sboxes, Goal).	
rewrite_each(List, Quant, Expr, ExpandedList, Defs, Objects, Sboxes, Goal) :-
	must_be(List, proper_list, Goal, 0),
	rewrite_each1(List, Quant, Expr, ExpandedList, Defs, Objects, Sboxes, Goal).

rewrite_each1([], _, _, [], _, _, _, _).
rewrite_each1([X|Xs], Q, E, [Y2|Ys], Defs, Objects, Sboxes, Goal) :-
	copy_term(Q-E, X-Y1),
	rewrite_expr(Y1, Y2, Defs, Objects, Sboxes, Goal),
	rewrite_each1(Xs, Q, E, Ys, Defs, Objects, Sboxes, Goal).

% quick & dirty
select_objects([], _, []).
select_objects([Oid|Oids], L1, [X|L3]) :-
	X = object(Oid,_,_,_),
	select(X, L1, L2), !,
	select_objects(Oids, L2, L3).

% quick & dirty
select_sboxes([], _, [], _).
select_sboxes([Expr|Sids], L1, L3, Objects) :-
	Expr = Obj^sid,
	Obj = object(_,Sid,_,_),
	(   integer(Sid) ->
	    select_sboxes(L1, Sid, L2, L3, L4)
	;   memberchk(Obj, Objects),
	    fd_min(Sid, Min),
	    select_pboxes(L1, Min, L2, Obj, 0, L3, L4)
	),
	select_sboxes(Sids, L2, L4, Objects).

select_sboxes([], _, []) --> [].
select_sboxes([X|L1], Y, [X|L2]) -->
	{\+(X=sbox(Y,_,_,_))}, !,
	select_sboxes(L1, Y, L2).
select_sboxes([X|L1], Y, L2) --> [X],
	select_sboxes(L1, Y, L2).

select_pboxes([], _, [], _, _) --> [].
select_pboxes([X|L1], Y, [X|L2], Obj, I) -->
	{\+(X=sbox(Y,_,_,_))}, !,
	select_pboxes(L1, Y, L2, Obj, I).
select_pboxes([_|L1], Y, L2, Obj, I) --> [Obj^pbox(J)],
	{J is I+1},
	select_pboxes(L1, Y, L2, Obj, J).

select_jth_sbox_atts([], _, _, _, []).
select_jth_sbox_atts([Sid|Sids], J, A, Sboxes, [V|Values]) :-
	Sbox = sbox(Sid,_,_,_),
	select_jth_sbox(J, Sbox, Sboxes, Sboxes1),
	get_path(Sbox, A, V, [], []),
	integer(V),
	select_jth_sbox_atts(Sids, J, A, Sboxes1, Values).

select_jth_sbox(1, Sbox, Sboxes0, Sboxes) :-
	select(Sbox, Sboxes0, Sboxes), !.
select_jth_sbox(I, Sbox, Sboxes0, Sboxes) :-
	select(Sbox, Sboxes0, Sboxes1), !,
	J is I-1,
	select_jth_sbox(J, Sbox, Sboxes1, Sboxes).

conjuncts(true) --> !.
conjuncts(P #/\ Q) --> !,
	conjuncts(P),
	conjuncts(Q).
conjuncts(P) --> [P].

andify([], true).
andify([P|Ps], Conj) :-
	andify(Ps, P, Conj).

andify([], P, P).
andify([P|Ps], Q, Conj) :-
	andify(Ps, Q#/\P, Conj).

orify([], true).
orify([P|Ps], Disj) :-
	orify(Ps, P, Disj).

orify([], P, P).
orify([P|Ps], Q, Disj) :-
	orify(Ps, Q#\/P, Disj).

cardify(L1, U1, N, List, Expanded) :-
	cardify_filter(List, List1, L1, L2, U1, U2, N, N2),
	L3 is max(L2,0),
	U3 is min(U2,N2),
	cardify2(L3, U3, N2, List1, Expanded).

cardify_filter([], [], L, L, U, U, N, N).
cardify_filter([true|List], List1, L, L2, U, U2, N, N2) :- !,
	L1 is L-1,
	U1 is U-1,
	N1 is N-1,
	cardify_filter(List, List1, L1, L2, U1, U2, N1, N2).
cardify_filter([false|List], List1, L, L2, U, U2, N, N2) :- !,
	U1 is U-1,
	N1 is N-1,
	cardify_filter(List, List1, L, L2, U1, U2, N1, N2).
cardify_filter([X|List], [X|List1], L, L2, U, U2, N, N2) :-
	cardify_filter(List, List1, L, L2, U, U2, N, N2).

cardify2(L, U, _N, _Ps, false) :- L>U, !.
cardify2(0, N, N, _Ps, true) :- !.
cardify2(1, N, N, Ps, Exp) :- !,
	orify(Ps, Exp).
cardify2(N, N, N, Ps, Exp) :- !,
	andify(Ps, Exp).
cardify2(0, 0, _N, Ps, #\ Exp) :- !,
	orify(Ps, Exp).
cardify2(0, N1, N, Ps, #\ Exp) :- N1=:=N-1, !,
	andify(Ps, Exp).
cardify2(L, U, N, [P|Ps], (P #/\ Exp1 #\/ #\P #/\ Exp2)) :-
	L1 is L-1,
	U1 is U-1,
	N1 is N-1,
	cardify2(L1, U1, N1, Ps, Exp1),
	cardify2(L , U , N1, Ps, Exp2).

arithop(A #= B, #=, A, B).
arithop(A #\= B, #\=, A, B).
arithop(A #< B, #<, A, B).
arithop(A #=< B, #=<, A, B).
arithop(A #> B, #>, A, B).
arithop(A #>= B, #>=, A, B).

evalop(#=, A, B, true)  :- A #= B, !.
evalop(#\=, A, B, true) :- A #\= B, !.
evalop(#<, A, B, true)  :- A #< B, !.
evalop(#=<, A, B, true) :- A #=< B, !.
evalop(#>, A, B, true)  :- A #> B, !.
evalop(#>=, A, B, true) :- A #>= B, !.
evalop(_, _, _, false).

get_path(object(OID,_,_,_), oid, OID, _Objects, _Sboxes) :- !.
get_path(object(_,SID,_,_), sid, SID, _Objects, _Sboxes) :- !.
get_path(object(_,_,X,_),   x(I),V,   _Objects, _Sboxes) :- !,
	nth1(I, X, V).
get_path(object(_,_,_,Atts),Key, V,   _Objects, _Sboxes) :- !,
	memberchk(Key-V, Atts).
get_path(Obj^pbox(J),      Attr, Val, Objects, Sboxes) :- !,
	get_pbox_path(Obj, J, Attr, Val, Objects, Sboxes).
get_path(sbox(SID,_,_,_),  sid, SID, _Objects, _Sboxes) :- !.
get_path(sbox(_,T,_,_),    t(I),V,   _Objects, _Sboxes) :- !,
	nth1(I, T, V).
get_path(sbox(_,_,L,_),    l(I),V,   _Objects, _Sboxes) :- !,
	nth1(I, L, V).
get_path(sbox(_,_,_,Atts), Key, V,   _Objects, _Sboxes) :- !,
	memberchk(Key-V, Atts).

get_pbox_path(Obj, J, Attr, Val, Objects, Sboxes) :-
	memberchk(Obj, Objects),
	Obj = object(_,Sid,_,_),
	fd_set(Sid, Set),
	fdset_to_list(Set, L),
	select_jth_sbox_atts(L, J, Attr, Sboxes, Values),
	same_values(Values, Val).

same_values([], _).
same_values([X|Xs], X) :-
	same_values(Xs, X).
	
% rewrite_away_not(Tree1, Tree2)
%   Eliminate -, *, /
rewrite_away_not(true, true) :- !.
rewrite_away_not(false, false) :- !.
rewrite_away_not(#\ X, Y) :- !,
	rewrite_away_not_neg(X, Y).
rewrite_away_not(X1 #/\ X2, Y1 #/\ Y2) :- !,
	rewrite_away_not(X1, Y1),
	rewrite_away_not(X2, Y2).
rewrite_away_not(X1 #\/ X2, Y1 #\/ Y2) :- !,
	rewrite_away_not(X1, Y1),
	rewrite_away_not(X2, Y2).
rewrite_away_not(R, R).

rewrite_away_not_neg(true, false) :- !.
rewrite_away_not_neg(false, true) :- !.
rewrite_away_not_neg(#\ X, Y) :- !,
	rewrite_away_not(X, Y).
rewrite_away_not_neg(X1 #/\ X2, Y1 #\/ Y2) :- !,
	rewrite_away_not_neg(X1, Y1),
	rewrite_away_not_neg(X2, Y2).
rewrite_away_not_neg(X1 #\/ X2, Y1 #/\ Y2) :- !,
	rewrite_away_not_neg(X1, Y1),
	rewrite_away_not_neg(X2, Y2).
rewrite_away_not_neg(A #= B, A #\= B).
rewrite_away_not_neg(A #\= B, A #= B).
rewrite_away_not_neg(A #< B, A #>= B).
rewrite_away_not_neg(A #=< B, A #> B).
rewrite_away_not_neg(A #> B, A #=< B).
rewrite_away_not_neg(A #>= B, A #< B).

% rewrite_relations(Tree1, Tree2)
%   Normalize all expressions into one of [Expr #>= 0, Expr #> 0]
rewrite_relations(true, true).
rewrite_relations(false, false).
rewrite_relations(X1 #/\ X2, Y1 #/\ Y2) :-
	rewrite_relations(X1, Y1),
	rewrite_relations(X2, Y2).
rewrite_relations(X1 #\/ X2, Y1 #\/ Y2) :-
	rewrite_relations(X1, Y1),
	rewrite_relations(X2, Y2).
rewrite_relations(X #= Y,  X-Y #>= 0 #/\ Y-X #>= 0).
rewrite_relations(X #\= Y, X-Y #>  0 #\/ Y-X #>  0).
rewrite_relations(X #< Y,  Y-X #>  0).
rewrite_relations(X #> Y,  X-Y #>  0).
rewrite_relations(X #=< Y, Y-X #>= 0).
rewrite_relations(X #>= Y, X-Y #>= 0).

% rewrite_away_ops(Tree1, Tree2)
%   Eliminate -, *, /
rewrite_away_ops(true, true) :- !.
rewrite_away_ops(false, false) :- !.
rewrite_away_ops(X1 #/\ X2, Y1 #/\ Y2) :- !,
	rewrite_away_ops(X1, Y1),
	rewrite_away_ops(X2, Y2).
rewrite_away_ops(X1 #\/ X2, Y1 #\/ Y2) :- !,
	rewrite_away_ops(X1, Y1),
	rewrite_away_ops(X2, Y2).
rewrite_away_ops(X #> 0,  Y #>  0) :- !,
	rewrite_away_ops(X, Y, rat(1,1)).
rewrite_away_ops(X #>= 0, Y #>= 0) :- !,
	rewrite_away_ops(X, Y, rat(1,1)).

rewrite_away_ops(min(X1,X2), Expr, M) :- !,
	rewrite_away_ops(X1, Y1, M),
	rewrite_away_ops(X2, Y2, M),
	(   rat_gez(M) -> Expr = min(Y1,Y2)
	;   Expr = max(Y1,Y2)
	).
rewrite_away_ops(max(X1,X2), Expr, M) :- !,
	rewrite_away_ops(X1, Y1, M),
	rewrite_away_ops(X2, Y2, M),
	(   rat_gez(M) -> Expr = max(Y1,Y2)
	;   Expr = min(Y1,Y2)
	).
rewrite_away_ops(X1 + X2, Y1 + Y2, M) :- !,
	rewrite_away_ops(X1, Y1, M),
	rewrite_away_ops(X2, Y2, M).
rewrite_away_ops(X1 - X2, Y1 + Y2, M) :- !,
	rewrite_away_ops(X1, Y1, M),
	rat_multiply(M, -1, M1),
	rewrite_away_ops(X2, Y2, M1).
rewrite_away_ops(X1 * X2, Y2, M) :-
	eval_ground(X1, V), !,
	rat_mul(M, V, M1),
	rewrite_away_ops(X2, Y2, M1).
rewrite_away_ops(X1 * X2, Y1, M) :- !,
	eval_ground(X2, V),
	rat_mul(M, V, M1),
	rewrite_away_ops(X1, Y1, M1).
rewrite_away_ops(X1 / X2, Y1, M) :- !,
	eval_ground(X2, V),
	rat_div(M, V, M1),
	rewrite_away_ops(X1, Y1, M1).
rewrite_away_ops(X, scaled(R,1), M) :-
	integer(X), !,
	rat_multiply(M, X, R).
rewrite_away_ops(X, scaled(M,X), M).

%   eval_ground(+Expr, -Rat)
eval_ground(X+Y, R) :- !,
	eval_ground(X, R1),
	eval_ground(Y, R2),
	rat_add(R1, R2, R).
eval_ground(X-Y, R) :- !,
	eval_ground(X, R1),
	eval_ground(Y, R2),
	rat_sub(R1, R2, R).
eval_ground(X*Y, R) :- !,
	eval_ground(X, R1),
	eval_ground(Y, R2),
	rat_mul(R1, R2, R).
eval_ground(X/Y, R) :- !,
	eval_ground(X, R1),
	eval_ground(Y, R2),
	rat_div(R1, R2, R).
eval_ground(min(X,Y), R) :- !,
	eval_ground(X, R1),
	eval_ground(Y, R2),
	rat_min_max(R1, R2, R, _).
eval_ground(max(X,Y), R) :- !,
	eval_ground(X, R1),
	eval_ground(Y, R2),
	rat_min_max(R1, R2, _, R).
eval_ground(X, rat(X,1)) :-
	integer(X).

% rewrite_away_min_max(Tree1, Tree2)
%   Eliminate min, max
rewrite_away_min_max(true, true).
rewrite_away_min_max(false, false).
rewrite_away_min_max(X1 #/\ X2, Y1 #/\ Y2) :-
	rewrite_away_min_max(X1, Y1),
	rewrite_away_min_max(X2, Y2).
rewrite_away_min_max(X1 #\/ X2, Y1 #\/ Y2) :-
	rewrite_away_min_max(X1, Y1),
	rewrite_away_min_max(X2, Y2).
rewrite_away_min_max(X #> 0,  Comb) :-
	rewrite_distribute_plus(X, Y),
	rewrite_min_max(Y, Comb, #>).
rewrite_away_min_max(X #>= 0, Comb) :-
	rewrite_distribute_plus(X, Y),
	rewrite_min_max(Y, Comb, #>=).

rewrite_min_max(min(X,Y), P1 #/\ P2, Rel) :- !,
	rewrite_min_max(X, P1, Rel),
	rewrite_min_max(Y, P2, Rel).
rewrite_min_max(max(X,Y), P1 #\/ P2, Rel) :- !,
	rewrite_min_max(X, P1, Rel),
	rewrite_min_max(Y, P2, Rel).
rewrite_min_max(X, P, Rel) :-
	P =.. [Rel,X,0].

rewrite_distribute_plus(min(X1,X2), min(Y1,Y2)) :-
	rewrite_distribute_plus(X1, Y1),
	rewrite_distribute_plus(X2, Y2).
rewrite_distribute_plus(max(X1,X2), max(Y1,Y2)) :-
	rewrite_distribute_plus(X1, Y1),
	rewrite_distribute_plus(X2, Y2).
rewrite_distribute_plus(X1+X2, Y) :-
	rewrite_distribute_plus(X1, Y1),
	rewrite_distribute_plus(X2, Y2),
	rewrite_distribute_plus(Y1, Y2, Y).
rewrite_distribute_plus(scaled(R,C), scaled(R,C)).

rewrite_distribute_plus(min(X1,X2), Y, min(Z1,Z2)) :- !,
	rewrite_distribute_plus(X1+Y, Z1),
	rewrite_distribute_plus(X2+Y, Z2).
rewrite_distribute_plus(Y, min(X1,X2), min(Z1,Z2)) :- !,
	rewrite_distribute_plus(Y+X1, Z1),
	rewrite_distribute_plus(Y+X2, Z2).
rewrite_distribute_plus(max(X1,X2), Y, max(Z1,Z2)) :- !,
	rewrite_distribute_plus(X1+Y, Z1),
	rewrite_distribute_plus(X2+Y, Z2).
rewrite_distribute_plus(Y, max(X1,X2), max(Z1,Z2)) :- !,
	rewrite_distribute_plus(Y+X1, Z1),
	rewrite_distribute_plus(Y+X2, Z2).
rewrite_distribute_plus(X, Y, X+Y).

% rewrite_linearize(Tree1, Tree2)
%   Normalize all expressions into one of [SP#>=C,true,false] where
%   SP is a list of Var-Coeff pairs, where Coeffs and C are integers.
rewrite_linearize(true, true).
rewrite_linearize(false, false).
rewrite_linearize(X1 #/\ X2, Y1 #/\ Y2) :-
	rewrite_linearize(X1, Y1),
	rewrite_linearize(X2, Y2).
rewrite_linearize(X1 #\/ X2, Y1 #\/ Y2) :-
	rewrite_linearize(X1, Y1),
	rewrite_linearize(X2, Y2).
rewrite_linearize(Expr #>= 0, Rel) :-
	rewrite_linearize(Expr, 0, Rel).
rewrite_linearize(Expr #> 0, Rel) :-
	rewrite_linearize(Expr, 1, Rel).

rewrite_linearize(Expr, Inc, Rel) :-
	linearize(Expr, Vector0, []),
	keysort(Vector0, Vector1),
	key_add_rat_values(Vector1, Vector2),
	keys_and_values(Vector2, Vars, Rats),
	rat_divisors(Rats, Divs),
	lcm(Divs, LCM),
	rats_to_ints(Rats, LCM, Ints),
	keys_and_values(Vector3, Vars, Ints), !,
	(   Vector3=[1-C|Vector4] -> RHS is Inc-C
	;   Vector4 = Vector3,
	    RHS = Inc
	),
	(   Vector4==[] -> evalop(#>=, 0, RHS, Rel)
	;   Rel = (Vector4 #>= RHS)
	).

linearize(X+Y) -->
	linearize(X),
	linearize(Y).
linearize(scaled(R,X)) --> [X-R].

key_add_rat_values([], []).
key_add_rat_values([K1-C1,K2-C2|V1], V3) :-
	K1==K2, !,
	rat_add(C1, C2, C3),
	(C3==rat(0,1) -> V2=V1 ; V2=[K1-C3|V1]),
	key_add_rat_values(V2, V3).
key_add_rat_values([KC|V1], [KC|V2]) :-
	key_add_rat_values(V1, V2).

rat_gez(rat(A,_)) :- A>=0.

rat_divisors([], []).
rat_divisors([rat(_,D)|L1], [D|L2]) :-
	rat_divisors(L1, L2).

rat_normalize(rat(A,B), rat(C,D)) :-
	G is gcd(A,B)*sign(B),
	C is A//G,
	D is B//G.

rat_multiply(rat(A,B), M, R) :-
	C is M*A,
	rat_normalize(rat(C,B), R).

rat_add(rat(A,B), rat(C,D), R) :-
	E is A*D+B*C,
	F is B*D,
	rat_normalize(rat(E,F), R).

rat_sub(rat(A,B), rat(C,D), R) :-
	E is A*D-B*C,
	F is B*D,
	rat_normalize(rat(E,F), R).

rat_mul(rat(A,B), rat(C,D), R) :-
	E is A*C,
	F is B*D,
	rat_normalize(rat(E,F), R).

rat_div(rat(A,B), rat(C,D), R) :-
	E is A*D,
	F is B*C,
	rat_normalize(rat(E,F), R).

rat_min_max(rat(A,B), rat(C,D), Min, Max) :-
	(   A*D =< B*C ->
	    Min=rat(A,B), Max=rat(C,D)
	;   Max=rat(A,B), Min=rat(C,D)
	).

rats_to_ints([], _, []).
rats_to_ints([rat(A,B)|L1], M, [X|L2]) :-
	X is A*M//B,
	rats_to_ints(L1, M, L2).

lcm(L, M) :-
	lcm(L, 1, M).

lcm([], M, M).
lcm([X|Xs], M0, M) :-
	M1 is X*M0//gcd(X,M0),
	lcm(Xs, M1, M).
	
% rewrite_simplify(Tree1, Tree2, Context)
%   Simplify away all occurrences of 'true' and 'false' except at the root.
rewrite_simplify(true, true, _).
rewrite_simplify(false, false, _).
rewrite_simplify(P #/\ Q, Rel, Context) :-
	rewrite_simplify(P, P1, Context),
	(   P1==false -> Rel = false
	;   rewrite_simplify(Q, Q1, Context),
	    (   P1==true -> Rel = Q1
	    ;   Q1==true -> Rel = P1
	    ;   Q1==false -> Rel = false
	    ;   Rel = (P1#/\Q1)
	    )
	).
rewrite_simplify(P #\/ Q, Rel, Context) :-
	rewrite_simplify(P, P1, Context),
	(   P1==true -> Rel = true
	;   rewrite_simplify(Q, Q1, Context),
	    (   P1==false -> Rel = Q1
	    ;   Q1==true -> Rel = true
	    ;   Q1==false -> Rel = P1
	    ;   Rel = (P1#\/Q1)
	    )
	).
rewrite_simplify(L#>=R, Rel, Context) :-
	linear_min_max(L, 0, Min, 0, Max, Context),
	(   Min>=R -> Rel = true
	;   Max<R  -> Rel = false
	;   Rel = (L#>=R)
	).

linear_min_max([], Min, Min, Max, Max, _).
linear_min_max([Ref-C|Linear], Min0, Min, Max0, Max, Context) :-
	ref_min_max(Ref, Rmin, Rmax, Context),
	(   C>0 ->
	    Min1 is Min0+C*Rmin, Max1 is Max0+C*Rmax
	;   Min1 is Min0+C*Rmax, Max1 is Max0+C*Rmin
	),
	linear_min_max(Linear, Min1, Min, Max1, Max, Context).

ref_min_max(object(ID,_,_,_)^x(J), Min, Max, Context) :- !,
	Context = context(Objects,_),
	memberchk(object(ID,_,Xs,_), Objects),
	nth1(J, Xs, X),
	fd_min(X, Min),
	fd_max(X, Max).
ref_min_max((object(ID,_,_,_)^pbox(J))^A, Min, Max, Context) :- !,
	Context = context(Objects,Sboxes),
	memberchk(object(ID,SID,_,_), Objects),
	fd_set(SID, S),
	fdset_to_list(S, L),
	select_jth_sbox_atts(L, J, A, Sboxes, Values),
	min_member(Min, Values),
	max_member(Max, Values).
	

% compile_linearized(Tree, K) --> list of: o(OID)-Procedure.

compile_linearized(Tree, K) -->
	{Tree\==fail},
	{conjuncts(Tree, Parts1, [])},
	{tag_by_objects(Parts1, Parts2)},
	{keysort(Parts2, Parts3)},
	{keyclumped(Parts3, Parts4)},
	compile_parts(Parts4, K).

tag_by_objects([], []).
tag_by_objects([Part|L1], [Objs-Part|L2]) :-
	objects_of_tree(Part, IDs, []),
	sort(IDs, Objs),
	tag_by_objects(L1, L2).

compile_parts([], _) --> [].
compile_parts([Objs-List|Parts], K) -->
	{andify(List, Tree1)},
	{commute_tree(Tree1, Tree2, _)},
	compile_tree_for_objects(Objs, Tree2, K),
	compile_parts(Parts, K).

compile_tree_for_objects([], _, _) --> [].
compile_tree_for_objects([ID|IDs], Tree, K) -->
	{forbidden_set_for_object(Tree, ID, RHS)},
	emit(RHS, o(ID)),
	compile_tree_for_objects(IDs, Tree, K).

objects_of_tree(R1 #/\ R2) -->
	objects_of_tree(R1),
	objects_of_tree(R2).
objects_of_tree(R1 #\/ R2) -->
	objects_of_tree(R1),
	objects_of_tree(R2).
objects_of_tree(L #>= _) -->
	objects_of_vector(L).

objects_of_vector([]) --> [].
objects_of_vector([object(ID,_,_,_)^_-_|L]) --> !, [ID],
	objects_of_vector(L).	
objects_of_vector([(object(ID,_,_,_)^pbox(_))^_-_|L]) --> !, [ID],
	objects_of_vector(L).	

commute_tree(P #\/ Q, Tree, Depth) :- !,
	commute_tree(P, P1, D1),
	commute_tree(Q, Q1, D2),
	Depth is D1+D2+1,
	(D1=<D2 -> Tree = (P1#\/Q1) ; Tree = (Q1#\/P1)).
commute_tree(P #/\ Q, Tree, Depth) :- !,
	commute_tree(P, P1, D1),
	commute_tree(Q, Q1, D2),
	Depth is D1+D2+1,
	(D1=<D2 -> Tree = (P1#/\Q1) ; Tree = (Q1#/\P1)).
commute_tree(Tree, Tree, 0).

forbidden_set_for_object(P #\/ Q, ID, SExpr) :-
	forbidden_set_for_object(P, ID, SExpr1),
	forbidden_set_for_object(Q, ID, SExpr2),
	sexpr_intersection(SExpr1, SExpr2, SExpr).
forbidden_set_for_object(P #/\ Q, ID, SExpr) :-
	forbidden_set_for_object(P, ID, SExpr1),
	forbidden_set_for_object(Q, ID, SExpr2),
	sexpr_union(SExpr1, SExpr2, SExpr).
forbidden_set_for_object(L#>=R, ID, SExpr) :-
	findall(SExpr1, build_an_ix(ID,L,R,SExpr1), SExprs),
	SExprs\==[], !,
	sexprs_union(SExprs, SExpr).
forbidden_set_for_object(L#>=R, _, gez(Sum)) :-
	R1 is -R,
	build_ix_sum(L, 1, max, int(R1), Sum).

build_an_ix(ID, L, R, SExpr) :- 
	select(Ref-C, L, L1),
	build_ix(Ref, ID, C, L1, R, SExpr).

build_ix(Ref, Oid, C, L1, R, {OExpr}) :- 
	Ref = object(Oid,_,Orig,_)^x(J),
	C>0, !,
	R1 is R-C,
	build_ix_sum(L1, -1, min, int(R1), Sum),
	(   C =:= 1 -> Quot = Sum
	;   Quot = ceildiv(Sum,C)
	),
	length(Orig, K),
	build_oexpr(0, J, K, inf..Quot, OExpr).
build_ix(Ref, Oid, C, L1, R, SExpr) :-
	Ref = (object(Oid,_,_,_)^pbox(J))^A,
	C>0, !,
	R1 is R-C,
	build_ix_sum(L1, -1, min, int(R1), Sum),
	(   C =:= 1 -> Quot = Sum
	;   Quot = ceildiv(Sum,C)
	),
	SExpr = sids(Oid,J,A,=<,Quot).
build_ix(Ref, Oid, C, L1, R, {OExpr}) :-
	Ref = object(Oid,_,Orig,_)^x(J), !,
	R1 is -R-C,
	build_ix_sum(L1, 1, max, int(R1), Sum),
	(   C =:= -1 -> Quot = Sum
	;   D is -C, Quot = floordiv(Sum,D)
	),
	length(Orig, K),
	build_oexpr(0, J, K, Quot..sup, OExpr).
build_ix(Ref, Oid, C, L1, R, SExpr) :-
	Ref = (object(Oid,_,_,_)^pbox(J))^A,
	R1 is -R-C,
	build_ix_sum(L1, 1, max, int(R1), Sum),
	(   C =:= -1 -> Quot = Sum
	;   D is -C, Quot = floordiv(Sum,D)
	),
	SExpr = sids(Oid,J,A,>=,Quot).

build_ix_sum([], _, _, Sum, Sum).
build_ix_sum([Ref-Con|V], Sign, Tag, Sum0, Sum) :-
	SCon is Sign*Con,
	SCon>0, !,
	build_term(Ref, Tag, Term),
	(   SCon =:= 1 -> Sum1 = Sum0+Term
	;   Sum1 = Sum0+mul(Term,SCon)
	),
	build_ix_sum(V, Sign, Tag, Sum1, Sum).
build_ix_sum([Ref-Con|V], Sign, Tag, Sum0, Sum) :-
	ACon is -Sign*Con,
	flip_tag(Tag, Fag),
	build_term(Ref, Fag, Term),
	(   ACon =:= 1 -> Sum1 = Sum0-Term
	;   Sum1 = Sum0-mul(Term,ACon)
	),
	build_ix_sum(V, Sign, Tag, Sum1, Sum).

build_term(object(ID,_,_,_)^x(J),        min, xmin(ID,J)) :- !.
build_term(object(ID,_,_,_)^x(J),        max, xmax(ID,J)) :- !.
build_term((object(ID,_,_,_)^pbox(N))^A, min, pmin(ID,N,A)) :- !.
build_term((object(ID,_,_,_)^pbox(N))^A, max, pmax(ID,N,A)) :- !.
build_term(X, _, X).

flip_tag(min, max).
flip_tag(max, min).

build_oexpr(I, _, K, _, []) :-
	I > K, !.
build_oexpr(I, Key, K, Range, [Range|OExpr]) :-
	J is I+1,
	J=:=Key, !,
	build_oexpr(J, Key, K, Range, OExpr).
build_oexpr(I, Key, K, Range, [inf..sup|OExpr]) :- !,
	J is I+1,
	build_oexpr(J, Key, K, Range, OExpr).

%% union and intersection

sexprs_union([SE], SE) :- !.
sexprs_union([SE0|SEs], SE) :-
	sexprs_union(SEs, SE0, SE).

sexprs_union([], SE, SE).
sexprs_union([SE1|SEs], SE0, SE) :-
	sexpr_union(SE0, SE1, SE2),
	sexprs_union(SEs, SE2, SE).

sexpr_union(SE1, SE2, SE1) :-
	sexpr_subset(SE2, SE1), !.
sexpr_union(SE1, SE2, SE2) :-
	sexpr_subset(SE1, SE2), !.
sexpr_union({OE1}, {OE2}, {OE3}) :-
	oexpr_union(OE1, OE2, OE3, 0), !.
sexpr_union(SE1, SE2, SE1\/SE2).

% true if at most one pair of ranges is not the same,
% and that one must overlap
oexpr_union([], [], [], _).
oexpr_union([R|OE1], [R|OE2], [R|OE3], K) :- !,
	oexpr_union(OE1, OE2, OE3, K).
oexpr_union([A1..B1|OE1], [A2..B2|OE2], [A3..B3|OE3], 0) :-
	intervals_overlap(A1, B1, A2, B2),
	bound_min(A1, A2, A3),
	bound_max(B1, B2, B3),
	oexpr_union(OE1, OE2, OE3, 1).

intervals_overlap(A1, B1, A2, B2) :-
	simple_aexpr(A1, _, A1v),
	simple_aexpr(A2, _, A2v),
	simple_aexpr(B1, _, B1v),
	simple_aexpr(B2, _, B2v),
	A1v=<B2v,
	A2v=<B1v.

sexpr_intersection(SE1, SE2, SE1) :-
	sexpr_subset(SE1, SE2), !.
sexpr_intersection(SE1, SE2, SE2) :-
	sexpr_subset(SE2, SE1), !.
sexpr_intersection({OE1}, {OE2}, SE3) :- !,
	(oexpr_intersection(OE1, OE2, OE3) -> SE3 = {OE3} ; SE3 = {}).
% sexpr_intersection(SE1\/SE2, SE3, SE6) :- !, % maintain DNF
% 	sexpr_intersection(SE1, SE3, SE4),
% 	sexpr_intersection(SE2, SE3, SE5),
% 	sexpr_union(SE4, SE5, SE6).
% sexpr_intersection(SE1, SE2\/SE3, SE6) :- !, % maintain DNF
% 	sexpr_intersection(SE1, SE2, SE4),
% 	sexpr_intersection(SE1, SE3, SE5),
% 	sexpr_union(SE4, SE5, SE6).
sexpr_intersection(SE1, SE2, SE1/\SE2).

oexpr_intersection([], [], []).
oexpr_intersection([A1..B1|OE1], [A2..B2|OE2], [A3..B3|OE3]) :-
	bound_max(A1, A2, A3),
	bound_min(B1, B2, B3),
	\+aexpr_greater_than(A3, B3),
	oexpr_intersection(OE1, OE2, OE3).

sexpr_subset({}, _) :- !.
sexpr_subset(_, {}) :- !, fail.
sexpr_subset(gez(_), _) :- !, fail.
sexpr_subset(_, gez(_)) :- !, fail.
sexpr_subset(sids(_,_,_,_,_), _) :- !, fail.
sexpr_subset(_, sids(_,_,_,_,_)) :- !, fail.
sexpr_subset(SE1/\SE2, SE3) :- !,
	(   sexpr_subset(SE1, SE3)
	;   sexpr_subset(SE2, SE3)
	).
sexpr_subset(SE1\/SE2, SE3) :- !,
	sexpr_subset(SE1, SE3),
	sexpr_subset(SE2, SE3).
sexpr_subset(SE1, SE2/\SE3) :- !,
	sexpr_subset(SE1, SE2),
	sexpr_subset(SE1, SE3).
sexpr_subset(SE1, SE2\/SE3) :- !,
	(   sexpr_subset(SE1, SE2)
	;   sexpr_subset(SE1, SE3)
	).
sexpr_subset({OE1}, {OE2}) :-
	oexpr_subset(OE1, OE2).

oexpr_subset([], []).
oexpr_subset([A1..B1|OE1], [A2..B2|OE2]) :-
	bound_min(A1, A2, A2),
	bound_max(B1, B2, B2),
	oexpr_subset(OE1, OE2).

aexpr_greater_than(sup, _) :- !.
aexpr_greater_than(inf, _) :- !, fail.
aexpr_greater_than(_, sup) :- !, fail.
aexpr_greater_than(_, inf) :- !.
aexpr_greater_than(int(X), int(Y)) :- !,
	X>Y.
aexpr_greater_than(Min, Max) :-
	decompose_sum(Min, Rmax, C),
	decompose_sum(Max, Rmin, D),
	C>D,
	ixref(Rmax, R, max),
	ixref(Rmin, R, min).

bound_min(inf, _, inf) :- !.
bound_min(_, inf, inf) :- !.
bound_min(sup, X, X) :- !.
bound_min(X, sup, X) :- !.
bound_min(int(X), int(Y), int(Z)) :- !,
	Z is min(X,Y).
bound_min(X, Y, Expr) :-
	decompose_sum(X, Var, Con1),
	decompose_sum(Y, Var, Con2), !,
	(Con1=<Con2 -> Expr=X ; Expr=Y).
bound_min(X, Y, min(X,Y)).

bound_max(sup, _, sup) :- !.
bound_max(_, sup, sup) :- !.
bound_max(inf, X, X) :- !.
bound_max(X, inf, X) :- !.
bound_max(int(X), int(Y), int(Z)) :- !,
	Z is max(X,Y).
bound_max(X, Y, Expr) :-
	decompose_sum(X, Var, Con1),
	decompose_sum(Y, Var, Con2), !,
	(Con1>=Con2 -> Expr=X ; Expr=Y).
bound_max(X, Y, max(X,Y)).

decompose_sum(int(C) + X, X, C) :- !,
	ixref(X, _, _).
decompose_sum(X + int(C), X, C) :- !,
	ixref(X, _, _).
decompose_sum(X - int(C), X, D) :- !,
	D is -C,
	ixref(X, _, _).
decompose_sum(X, X, 0) :-
	ixref(X, _, _).

ixref(xmin(A,B), x(A,B), min).
ixref(xmax(A,B), x(A,B), max).
ixref(pmin(A,B,C), p(A,B,C), min).
ixref(pmax(A,B,C), p(A,B,C), max).

%%% S-expression to procedure

emit({}, _) --> !.
% emit(SE1\/SE2, X) --> !, % hinders common subexpression elimination
% 	emit(SE1, X),
% 	emit(SE2, X).
emit(SE, X) --> [X-Proc],
	{sexpr_to_procedure(SE, Proc)}.

sexpr_to_procedure(SE, procedure(NTemp,Dest,Insns,Dep2)) :-
	empty_avl(Map1),
	sexpr_to_procedure(SE, Dest, Map1, _, Insns, []),
	ascending_values(Insns, 0, NTemp),
	dependencies(Insns, Dep1, []),
	sort(Dep1, Dep2).

sexpr_to_procedure(SExpr, Dest, Map, Map) -->
	{avl_fetch(SExpr, Map, Dest)}, !.
sexpr_to_procedure(SExpr, Dest, Map1, Map3) -->
	sexpr_to_procedure1(SExpr, Dest, Map1, Map2),
	{avl_store(SExpr, Map2, Dest, Map3)}.

sexpr_to_procedure1({OExpr}, Dest, Map1, Map2) -->
	oexpr_to_procedure(OExpr, Map1, Map2, Args),
	[obox(Args)-Dest].
sexpr_to_procedure1(gez(A), Dest, Map1, Map2) -->
	aexpr_to_procedure(A, Ad, Map1, Map2),
	[gez(Ad)-Dest].
sexpr_to_procedure1(sids(A,B,C,D,E), Dest, Map1, Map2) -->
	aexpr_to_procedure(E, Ed, Map1, Map2),
	[sids(A,B,C,D,Ed)-Dest].
sexpr_to_procedure1(SE1/\SE2, Dest3, Map1, Map2) -->
	sexpr_to_procedure(SE1, Dest1, Map1, Map2),
	{sexpr_to_procedure(SE2, Dest2, Map2, _, Sub, [])},
	[intersection(Dest1,Dest2,Sub)-Dest3].
sexpr_to_procedure1(SE1\/SE2, Dest3, Map1, Map2) -->
	sexpr_to_procedure(SE1, Dest1, Map1, Map2),
	{sexpr_to_procedure(SE2, Dest2, Map2, _, Sub, [])},
	[union(Dest1,Dest2,Sub)-Dest3].

oexpr_to_procedure([], Map, Map, []) --> [].
oexpr_to_procedure([A..B|OExpr], Map1, Map4, [Ad..Bd|Dest]) -->
	aexpr_to_procedure(A, Ad, Map1, Map2),
	aexpr_to_procedure(B, Bd, Map2, Map3),
	oexpr_to_procedure(OExpr, Map3, Map4, Dest).

aexpr_to_procedure(Aexpr, Dest, Map, Map) -->
	{avl_fetch(Aexpr, Map, Dest)}, !.
aexpr_to_procedure(Aexpr, Dest, Map1, Map3) -->
	aexpr_to_procedure1(Aexpr, Dest, Map1, Map2),
	{avl_store(Aexpr, Map2, Dest, Map3)}.

aexpr_to_procedure1(inf, Dest, Map, Map) --> [inf-Dest].
aexpr_to_procedure1(sup, Dest, Map, Map) --> [sup-Dest].
aexpr_to_procedure1(int(I), Dest, Map, Map) --> [int(I)-Dest].
aexpr_to_procedure1(xmin(I,J), Dest, Map, Map) --> [xmin(I,J)-Dest].
aexpr_to_procedure1(xmax(I,J), Dest, Map, Map) --> [xmax(I,J)-Dest].
aexpr_to_procedure1(pmin(I,J,A), Dest, Map, Map) --> [pmin(I,J,A)-Dest].
aexpr_to_procedure1(pmax(I,J,A), Dest, Map, Map) --> [pmax(I,J,A)-Dest].
aexpr_to_procedure1(+(AE1,AE2), Dest3, Map1, Map3) -->
	aexpr_to_procedure(AE1, Dest1, Map1, Map2),
	aexpr_to_procedure(AE2, Dest2, Map2, Map3),
	[plus(Dest1,Dest2)-Dest3].
aexpr_to_procedure1(-(AE1,AE2), Dest3, Map1, Map3) -->
	aexpr_to_procedure(AE1, Dest1, Map1, Map2),
	aexpr_to_procedure(AE2, Dest2, Map2, Map3),
	[minus(Dest1,Dest2)-Dest3].
aexpr_to_procedure1(mul(AE,Int), Dest2, Map1, Map2) -->
	aexpr_to_procedure(AE, Dest1, Map1, Map2),
	[mul(Dest1,Int)-Dest2].
aexpr_to_procedure1(floordiv(AE,Int), Dest2, Map1, Map2) -->
	aexpr_to_procedure(AE, Dest1, Map1, Map2),
	[floordiv(Dest1,Int)-Dest2].
aexpr_to_procedure1(ceildiv(AE,Int), Dest2, Map1, Map2) -->
	aexpr_to_procedure(AE, Dest1, Map1, Map2),
	[ceildiv(Dest1,Int)-Dest2].

ascending_values([], NTemp, NTemp).
ascending_values([intersection(_,_,Sub)-K|Insns], I, L) :- !,
	ascending_values(Sub, I, J),
	K is J+1,
	ascending_values(Insns, K, L).
ascending_values([union(_,_,Sub)-K|Insns], I, L) :- !,
	ascending_values(Sub, I, J),
	K is J+1,
	ascending_values(Insns, K, L).
ascending_values([_-J|Insns], I, K) :-
	J is I+1,
	ascending_values(Insns, J, K).

dependencies([]) --> [].
dependencies([Insn-_|Insns]) -->
	dependency(Insn),
	dependencies(Insns).

dependency(xmin(ID,_)) --> !, [o(ID)].
dependency(xmax(ID,_)) --> !, [o(ID)].
dependency(pmin(ID,_,_)) --> !, [o(ID)].
dependency(pmax(ID,_,_)) --> !, [o(ID)].
% dependency(sids(ID,_,_,_,_)) --> !, [o(ID)].
dependency(intersection(_,_,Insns)) --> !,
	dependencies(Insns).
dependency(union(_,_,Insns)) --> !,
	dependencies(Insns).
dependency(_) --> [].

all_dependencies([]) --> [].
all_dependencies([Target-procedure(_,_,_,Srcs)|L]) -->
	all_dependencies(Srcs, Target),
	all_dependencies(L).

all_dependencies([], _) --> [].
all_dependencies([X|Xs], Y) --> [X-Y],
	all_dependencies(Xs, Y).

codelength([], N, N).
codelength([Insn-_|Insns], I, L) :-
	J is I+1,
	insnlength(Insn, J, K),
	codelength(Insns, K, L).

insnlength(intersection(_,_,Insns), I, K) :- !,
	codelength(Insns, I, K).
insnlength(union(_,_,Insns), I, K) :- !,
	codelength(Insns, I, K).
insnlength(_, I, K) :-
	K is I+1.

total_codelength([], K, K, N, N).
total_codelength([_-procedure(_,_,Insns,_)|List], I, K, L, N) :-
	codelength(Insns, I, J),
	M is L+1,
	total_codelength(List, J, K, M, N).
