:-use_module(library(lists)).
:-use_module(library(clpfd)).
:-ensure_loaded(geosr).


variables(Boxes, Container, [S1|Ss], Vars) :-
	container(1, _, W, H, L),
	mkobject(0, [0,0,0], [W,H,L], [], Container, S1),
	mkobjects(0, 9, [W,H,L], Boxes, Ss, Origs),
	append(Origs, Vars).

mkobject(ID, Orig, Dim, Atts,
	 object(ID,ID,Orig,Atts),
	 sbox(ID,[0,0,0],Dim)).

mkobjects(N, N, _, [], [], []) :- !.
mkobjects(I, N, UB, [O|Os], [S|Ss], [Origs|Origss]) :-
	J is I+1,
	object(J, _, L0, L1, L2, W),
	upper_bounds(UB, Origs),
	mkobject(J, Origs, [L0,L1,L2], [weight-W], O, S),
	mkobjects(J, N, UB, Os, Ss, Origss).

upper_bounds([], []).
upper_bounds([U|Us], [X|Xs]) :-
	X in 0..U,
	upper_bounds(Us, Xs).

problem(Boxes) :-
	variables(Boxes, Container, Sboxes, Vars),
	rule(lib, R0, R1),
	rule(inside, R1, R2),
	rule(non_overlap, R2, R3),
	rule(gravity, R3, R4),
	rule(stack_weight, R4, R5),
	rule(stack_oversize, R5, R6),
	rule(wedging, R6, R7),
	R7 = [forall(Box,Boxes,
		     inside(Box,Container) #/\
		     gravity(Box,Boxes) #/\
	             % wedging(Box,Container,Boxes) #/\
		     true),
	      forall(Box1,Boxes,
		     forall(Box2,Boxes,
			    non_overlap(Box1,Box2) #/\
			    stack_weight(Box1,Box2) #/\
			    stack_oversize(Box1,Box2) #/\
			    true))
	     ],
	geosr([Container|Boxes], Sboxes, R0),
	garbage_collect, statistics,
	labeling([], Vars),
	garbage_collect, statistics.

rule(lib) -->
	[(origin(O1,S1,D) ---> O1^x(D)+S1^t(D))],
	[(end(O1,S1,D) ---> O1^x(D)+S1^t(D)+S1^l(D))],
	[(soverlap(O1,O2,S1,S2,D) --->  % sboxes overlap in dim D
	  end(O1,S1,D) #> origin(O2,S2,D) #/\
	  end(O2,S2,D) #> origin(O1,S1,D))],
	[(oversize(O1,O2,S1,S2,D) ---> % amount by which two sboxes overlap in dim D
	  max(max(origin(O1,S1,D),origin(O2,S2,D)) - min(origin(O1,S1,D),origin(O2,S2,D)),
	      max(end(O1,S1,D),   end(O2,S2,D)) -    min(end(O1,S1,D),   end(O2,S2,D))))].	 
rule(inside) -->
	[(inside(O1,O2) ---> % O1 non-strictly inside O2 in all dims
	  forall(S1,sboxes([O1^sid]),
		 exists(S2,sboxes([O2^sid]),
			origin(O2,S2,1) #=< origin(O1,S1,1) #/\
			end(O1,S1,1) #=< end(O2,S2,1) #/\
			origin(O2,S2,2) #=< origin(O1,S1,2) #/\
			end(O1,S1,2) #=< end(O2,S2,2) #/\
			origin(O2,S2,3) #=< origin(O1,S1,3) #/\
			end(O1,S1,3) #=< end(O2,S2,3))))].
rule(non_overlap) --> 
	[(non_overlap(O1,O2) ---> % O1 does not overlap O2 in all dims
	  O1^oid #< O2^oid #=>
	  forall(S1,sboxes([O1^sid]),
		 forall(S2,sboxes([O2^sid]),
			#\ soverlap(O1,O2,S1,S2,1) #\/
			#\ soverlap(O1,O2,S1,S2,2) #\/
			#\ soverlap(O1,O2,S1,S2,3))))].
rule(gravity) -->
	[(gravity(O1,Os) ---> % O1 is either on the floor on sitting on some other box
	  forall(S1,sboxes([O1^sid]),
		 (origin(O1,S1,3) #= 0 #\/
		  exists(O2,Os,
			 O1^oid#\=O2^oid #/\
			 exists(S2,sboxes([O2^sid]),
			        soverlap(O1,O2,S1,S2,1) #/\
			        soverlap(O1,O2,S1,S2,2) #/\
				origin(O1,S1,3) #= end(O2,S2,3))))))].
rule(stack_weight) -->
	[(stack_weight(O1,O2) ---> % O1 heavier than O2 => O1 not piled above O2
	  O1^weight #> O2^weight #=>
	  forall(S1,sboxes([O1^sid]),
		 forall(S2,sboxes([O2^sid]),
			origin(O1,S1,3) #>= end(O2,S2,3) #=>
			#\ soverlap(O1,O2,S1,S2,1) #\/
			#\ soverlap(O1,O2,S1,S2,2))))].
rule(stack_oversize) -->
	[(stack_oversize(O1,O2) ---> % for any two objects in a pile, the overhang can be at most 10
	  O1^oid#\=O2^oid #=>
	  forall(S1,sboxes([O1^sid]),
		 forall(S2,sboxes([O2^sid]),
			(soverlap(O1,O2,S1,S2,1) #/\
			 soverlap(O1,O2,S1,S2,2)) #=>
			(oversize(O1,O2,S1,S2,1) #=< 10 #/\
			 oversize(O1,O2,S1,S2,2) #=< 10))))].
rule(wedging) --> % based on Julien's formulation
	[(wedged(O1,S1,Oc,Sc,Os,D) ---> % all four faces of O1 in X/Y
				              % must lean against the container or against some other box
	  (origin(O1,S1,D) #= origin(Oc,Sc,D) #\/
	   exists(O2,Os,
		  O1^oid#\=O2^oid #/\
		  exists(S2,sboxes([O2^sid]),
		         origin(O1,S1,D) #= end(O2,S2,D)))) #/\
	  (end(O1,S1,D) #= end(Oc,Sc,D) #\/
	   exists(O2,Os,
		  O1^oid#\=O2^oid #/\
		  exists(S2,sboxes([O2^sid]),
			 end(O1,S1,D) #= origin(O2,S2,D)))))],

	[(wedging(O1,Oc,Os) --->
	  exists(Sc,sboxes([Oc^sid]),
		 forall(S1,sboxes([O1^sid]),
		        wedged(O1,S1,Oc,Sc,Os,1) #/\
		        wedged(O1,S1,Oc,Sc,Os,2))))].
		       

% PSA FEASIBLE INSTANCE (all rules posted except wedging_rule)

container(1, tc40, 1203, 235, 239).

object(1, cV8269, 224, 224, 111, 413).
object(2, pF8370, 224, 224, 74, 463).
object(3, cV8371, 224, 224, 74, 842).
object(4, cV8142, 224, 224, 148, 422).
object(5, cV8243, 224, 224, 111, 266).
object(6, cV8258, 224, 224, 111, 321).
object(7, pF8035, 224, 224, 222, 670).
object(8, pF7037, 155, 224, 222, 440).
object(9, cF6120, 112, 224, 148, 325).


/*
%%%%%%%%%%%%%%%%%%%%
% TOY FEASIBLE INSTANCE (all rules posted except weding_rule)

container(1, tc20, 6, 6, 3).
object(1, cp01, 1, 2, 1, 2).
object(2, cp01, 2, 1, 1, 5).
object(3, cp01, 2, 2, 2, 9).
object(4, cp01, 1, 2, 1, 4).
object(5, cp01, 3, 2, 1, 10).
object(6, cp01, 1, 1, 1, 2).
object(7, cp01, 3, 1, 2, 5).
object(8, cp01, 1, 2, 1, 3).
object(9, cp01, 1, 2, 1, 3).
object(10, cp01, 3, 2, 1, 7).
object(11, cp01, 3, 1, 1, 6).
object(12, cp01, 3, 2, 1, 5).
object(13, cp01, 2, 2, 1, 7).
object(14, cp01, 1, 3, 1, 6).
object(15, cp01, 2, 2, 1, 5).
object(16, cp01, 1, 2, 1, 3).

*/

/*
object(1, cV73_16, 155, 224, 74, 380).
object(2, c091_24, 155, 224, 86, 814).
object(3, c091_25, 155, 224, 86, 814).
object(4, d6_27, 117, 224, 109, 1141).
object(5, cV73_40, 155, 224, 74, 341).
object(6, cV73_41, 155, 224, 74, 300).
object(7, cV82_59, 224, 224, 111, 529).
object(8, pF73_60, 155, 224, 74, 290).
object(9, cV73_61, 155, 224, 74, 420).
object(10, cV73_65, 155, 224, 74, 606).
object(11, cF71_66, 155, 224, 148, 357).
object(12, pF72_77, 155, 224, 111, 320).
object(13, pF72_80, 155, 224, 111, 325).
*/

end_of_file.

Here is the first solution found (the boxes are in the same order as 
"objects" in the source file, i.e. 1,2,3,4,5,6,7,8,9) :

container([[0,1203,1203],[0,235,235],[0,239,239],[_22615]])

%object(n):  <X,Y,Z>

object(1) : <0,0,0>
object(2) : <224,0,0>
object(3) : <448,0,0>
object(4) : <224,0,74>
object(5) : <0,0,111>
object(6) : <448,0,74>
object(7) : <672,0,0>
object(8) : <896,0,0>
object(9) : <1051,0,0>

% REGRESSION TEST

| ?- problem(B).
% TOTAL #insns 50140 (was 56416 before merge)
% sweeping object 1
% found witness [0,0,0,1]
% found witness [979,11,128,1]
% sweeping object 2
% found witness [0,0,0,2]
% found witness [979,11,165,2]
% sweeping object 3
% found witness [0,0,0,3]
% found witness [979,11,165,3]
% sweeping object 4
% found witness [0,0,0,4]
% found witness [979,11,91,4]
% sweeping object 5
% found witness [0,0,0,5]
% found witness [979,11,128,5]
% sweeping object 6
% found witness [0,0,0,6]
% found witness [979,11,128,6]
% sweeping object 7
% found witness [0,0,0,7]
% found witness [979,11,0,7]
% sweeping object 8
% found witness [0,0,0,8]
% found witness [1048,11,0,8]
% sweeping object 9
% found witness [0,0,0,9]
% found witness [1091,11,91,9]
% sweeping object 1
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 1
% found witness [489,11,128,1]
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 1
% found witness [244,11,128,1]
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 1
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 1
% found witness [122,11,128,1]
% sweeping object 2
% sweeping object 3
% sweeping object 4
% found witness [224,0,0,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% found witness [224,0,0,7]
% sweeping object 8
% found witness [224,0,0,8]
% sweeping object 9
% sweeping object 1
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 1
% found witness [61,11,128,1]
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% found witness [224,0,0,9]
% sweeping object 1
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 1
% found witness [30,11,128,1]
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 1
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 1
% found witness [15,11,128,1]
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 1
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 1
% found witness [7,11,128,1]
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 1
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 1
% found witness [3,11,128,1]
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 1
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 1
% found witness [1,11,128,1]
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 1
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 1
% found witness [0,11,128,1]
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 1
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 1
% found witness [0,5,128,1]
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 1
% found witness [0,2,128,1]
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 1
% found witness [0,1,128,1]
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 1
% found witness [0,0,128,1]
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 1
% sweeping object 2
% found witness [224,0,0,2]
% sweeping object 3
% found witness [224,0,0,3]
% sweeping object 4
% sweeping object 5
% found witness [0,0,111,5]
% found witness [224,0,0,5]
% sweeping object 6
% found witness [0,0,111,6]
% found witness [224,0,0,6]
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 2
% sweeping object 2
% found witness [601,11,165,2]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 2
% found witness [412,11,165,2]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% found witness [448,0,0,7]
% sweeping object 8
% sweeping object 9
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 2
% found witness [318,11,165,2]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% found witness [448,0,0,8]
% sweeping object 9
% found witness [326,0,0,9]
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 2
% found witness [271,11,165,2]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% found witness [448,0,0,9]
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 2
% found witness [247,11,165,2]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 2
% found witness [235,11,165,2]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 2
% found witness [229,11,165,2]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 2
% found witness [226,11,165,2]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 2
% found witness [225,11,165,2]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 2
% found witness [224,11,165,2]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 2
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 2
% found witness [224,5,165,2]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 2
% found witness [224,2,165,2]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 2
% found witness [224,1,165,2]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 2
% found witness [224,0,165,2]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 2
% found witness [224,0,82,2]
% sweeping object 3
% sweeping object 4
% found witness [224,0,74,4]
% found witness [448,0,0,4]
% sweeping object 5
% found witness [448,0,0,5]
% sweeping object 6
% found witness [448,0,0,6]
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 2
% sweeping object 3
% found witness [448,0,0,3]
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 3
% found witness [713,11,165,3]
% sweeping object 4
% found witness [979,11,0,4]
% found witness [936,11,91,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% found witness [1091,11,0,9]
% found witness [936,11,91,9]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 3
% found witness [580,11,165,3]
% sweeping object 4
% found witness [803,11,91,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% found witness [672,0,0,7]
% sweeping object 8
% found witness [672,0,0,8]
% sweeping object 9
% found witness [803,11,91,9]
% sweeping object 3
% sweeping object 4
% found witness [803,11,91,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 3
% sweeping object 9
% sweeping object 4
% found witness [803,11,91,4]
% sweeping object 5
% sweeping object 6
% sweeping object 3
% found witness [514,11,165,3]
% sweeping object 4
% found witness [737,11,91,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% found witness [672,0,0,9]
% found witness [737,11,91,9]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 3
% found witness [481,11,165,3]
% sweeping object 4
% found witness [704,11,91,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% found witness [704,11,91,9]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 3
% found witness [464,11,165,3]
% sweeping object 4
% found witness [687,11,91,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% found witness [687,11,91,9]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 3
% found witness [456,11,165,3]
% sweeping object 4
% found witness [679,11,91,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% found witness [679,11,91,9]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 3
% found witness [452,11,165,3]
% sweeping object 4
% found witness [675,11,91,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% found witness [675,11,91,9]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 3
% found witness [450,11,165,3]
% sweeping object 4
% found witness [673,11,91,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% found witness [673,11,91,9]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 3
% found witness [449,11,165,3]
% sweeping object 4
% found witness [672,11,91,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% found witness [672,11,91,9]
% sweeping object 3
% sweeping object 4
% sweeping object 5
% sweeping object 3
% found witness [448,11,165,3]
% sweeping object 4
% found witness [671,11,91,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 3
% sweeping object 4
% found witness [671,11,91,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 3
% found witness [448,5,165,3]
% sweeping object 4
% found witness [671,11,91,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 3
% found witness [448,2,165,3]
% sweeping object 4
% found witness [671,11,91,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 3
% found witness [448,1,165,3]
% sweeping object 4
% found witness [671,11,91,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 3
% found witness [448,0,165,3]
% sweeping object 4
% found witness [671,10,91,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 3
% sweeping object 4
% sweeping object 5
% found witness [672,0,0,5]
% sweeping object 6
% found witness [672,0,0,6]
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 4
% sweeping object 5
% sweeping object 4
% found witness [458,10,74,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 4
% found witness [224,10,74,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 4
% sweeping object 5
% sweeping object 4
% found witness [224,5,74,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 4
% found witness [224,2,74,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 4
% found witness [224,1,74,4]
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 4
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 5
% found witness [224,0,74,5]
% found witness [458,10,128,5]
% sweeping object 6
% found witness [979,11,0,6]
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 5
% found witness [224,0,74,5]
% sweeping object 5
% found witness [0,10,128,5]
% sweeping object 6
% found witness [224,0,74,6]
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 5
% found witness [0,10,111,5]
% sweeping object 6
% found witness [448,0,74,6]
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 5
% found witness [0,5,111,5]
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 5
% found witness [0,2,111,5]
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 5
% found witness [0,1,111,5]
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 5
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 6
% found witness [713,11,0,6]
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 6
% found witness [458,10,74,6]
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 6
% found witness [453,10,74,6]
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 6
% found witness [450,10,74,6]
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 6
% found witness [449,10,74,6]
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 6
% found witness [448,10,74,6]
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 6
% found witness [448,5,74,6]
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 6
% found witness [448,2,74,6]
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 6
% found witness [448,1,74,6]
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 6
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 7
% found witness [824,11,0,7]
% sweeping object 8
% found witness [896,0,0,8]
% sweeping object 9
% sweeping object 7
% sweeping object 8
% sweeping object 7
% found witness [748,11,0,7]
% sweeping object 8
% sweeping object 9
% found witness [896,0,0,9]
% sweeping object 7
% sweeping object 8
% sweeping object 7
% sweeping object 9
% sweeping object 7
% found witness [710,11,0,7]
% sweeping object 8
% sweeping object 9
% sweeping object 7
% found witness [691,11,0,7]
% sweeping object 8
% sweeping object 9
% sweeping object 7
% found witness [681,11,0,7]
% sweeping object 8
% sweeping object 9
% sweeping object 7
% found witness [676,11,0,7]
% sweeping object 8
% sweeping object 9
% sweeping object 7
% found witness [674,11,0,7]
% sweeping object 8
% sweeping object 9
% sweeping object 7
% found witness [673,11,0,7]
% sweeping object 8
% sweeping object 9
% sweeping object 7
% found witness [672,11,0,7]
% sweeping object 8
% sweeping object 9
% sweeping object 7
% found witness [672,5,0,7]
% sweeping object 8
% sweeping object 9
% sweeping object 7
% found witness [672,2,0,7]
% sweeping object 8
% sweeping object 9
% sweeping object 7
% found witness [672,1,0,7]
% sweeping object 8
% sweeping object 9
% sweeping object 7
% sweeping object 8
% sweeping object 9
% sweeping object 8
% found witness [936,11,0,8]
% sweeping object 9
% found witness [1051,0,0,9]
% sweeping object 8
% sweeping object 8
% found witness [916,11,0,8]
% sweeping object 9
% sweeping object 8
% found witness [906,11,0,8]
% sweeping object 9
% sweeping object 8
% found witness [901,11,0,8]
% sweeping object 9
% sweeping object 8
% found witness [898,11,0,8]
% sweeping object 9
% sweeping object 8
% found witness [897,11,0,8]
% sweeping object 9
% sweeping object 8
% found witness [896,11,0,8]
% sweeping object 9
% sweeping object 8
% found witness [896,5,0,8]
% sweeping object 9
% sweeping object 8
% found witness [896,2,0,8]
% sweeping object 9
% sweeping object 8
% found witness [896,1,0,8]
% sweeping object 9
% sweeping object 8
% sweeping object 9
% sweeping object 9
% found witness [1071,11,0,9]
% sweeping object 9
% found witness [1061,11,0,9]
% sweeping object 9
% found witness [1056,11,0,9]
% sweeping object 9
% found witness [1053,11,0,9]
% sweeping object 9
% found witness [1052,11,0,9]
% sweeping object 9
% found witness [1051,11,0,9]
% sweeping object 9
% found witness [1051,5,0,9]
% sweeping object 9
% found witness [1051,2,0,9]
% sweeping object 9
% found witness [1051,1,0,9]
% sweeping object 9
B = [object(1,1,[0,0,0],[weight-413]),object(2,2,[224,0,0],[weight-463]),object(3,3,[448,0,0],[weight-842]),object(4,4,[224,0,74],[weight-422]),object(5,5,[0,0,111],[weight-266]),object(6,6,[448,0,74],[weight-321]),object(7,7,[672,0,0],[weight-670]),object(8,8,[896,0,0],[weight-440]),object(9,9,[1051,0,0],[weight-325])] ?
