:- module(r2cp, [rcp_var/2,
		 rcp_anavar/2,		 
		 rcp_variables/2,
		 rcp_variables_set/2,		 
		 rcp_att/3,
		 rcp_rec/2,
		 rcp_range/2,
		 variables/2,
		 variables_set/2,
		 max/3,
		 min/3,
		 log/3,
		 exp/3,
		 fd_lt/3,
		 fd_lt/2,
		 fd_leq/3,
		 fd_leq/2,
		 fd_eq/3,
		 fd_eq/2,
		 fd_geq/3,
		 fd_geq/2,
		 fd_gt/3,
		 fd_gt/2,
		 fd_neq/3,
		 fd_aeq/2,
		 fd_and/3,
		 fd_or/3,
		 fd_implies/3,
		 fd_equiv/3,
		 rcp_geost/3,
		 rcp_geost/4,
		 rcp_call/2,
		 rcp_assert/1, 
		 variable_ordering/1,
		 value_ordering/1,
		 rcp_heuristics/2,
		 rcp_labeling/1,
		 rcp_write/1,
		 rcp_flip/2,
		 rcp_rec_uid/2,
		 nth1/3,
		 set_interpret_mode/0,
		 init_interpreter/0,
		 interpret/2]).

:- use_module(library(lists)).
:- use_module(library(terms)).
:- use_module(library(avl)).
:- use_module(library(sets)).
:- use_module(library(clpfd)).
:- use_module(library(atts)).
:- use_module(library(samsort)).

:- use_module([subst]).
:- use_module(['./geosr/geosr.pl']).

:- dynamic rcp_rec_uid/2,
	   rcp_heuristics/2,
	   interpreter_mode/0.

:- attribute rcp_var_name/1.
verify_attributes(_, _, []).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% interpreter
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
init_interpreter :-
	init_rcp_rec_uid,
	counter_init(local_vars),
	counter_init(global_vars),
	retractall(rcp_heuristics(variable, _)),
	assert(rcp_heuristics(variable, [])),
	retractall(rcp_heuristics(value, _)),
	assert(rcp_heuristics(value, [])),
	retractall(rcp_heuristics(conjunct, _)),
	assert(rcp_heuristics(conjunct, [])),
	retractall(rcp_heuristics(disjunct, _)),
	assert(rcp_heuristics(disjunct, [])).

% prolog variable
interpret(Term, RTerm) :-
	var(Term),
	!,
	RTerm = Term.

% bond variable
interpret(var(Name, N), var(Name, N)) :-
	!.

% free (fd) variable
interpret(free_var(Name, I, N), R) :-
	!,
	rcp_var(from(Name, I, N), R).

% free (fd) variable
interpret(rcp_var(Name, V), rcp_var(Name, V)) :-
	!.

interpret(variables(Term), R) :-	
	interpret(Term, ITerm),
	variables(ITerm, R).

% fd variable lower bound
interpret(domain_min(Term), R) :-
	!,
	interpret(Term, ITerm),
	fd_min(ITerm, R).

% fd variable upper bound
interpret(domain_max(Term), R) :-
	!,
	interpret(Term, ITerm),
	fd_max(ITerm, R).

% fd variable domain size
interpret(domain_size(Term), R) :-
	!,
	interpret(Term, ITerm),
	fd_size(ITerm, R).

% integer
interpret(Term, RTerm) :-
	integer(Term),
	!,
	RTerm = Term.

interpret(min_integer, -268435456) :-
	!.

interpret(max_integer, 268435455) :-
	!.

% record
interpret(rcp_rec(AssocList), R) :-
	!,
	rcp_rec(AssocList, R).

% list
interpret([X], [Y]) :-
	var(X),
	!,
	X = Y.

interpret([X .. Y], R) :-
	!,
	interpret(X, L),
	interpret(Y, U),
	rcp_range([L .. U], R).

interpret(Term, Val) :-
	is_list(Term),
	!,
	interpret_list(Term, Val).

interpret(cons(A, B), '.'(IA, IB)) :-
	interpret(A, IA),
	interpret(B, IB).

interpret(length(Term), Val) :-
	!,
	interpret(Term, List),
	length(List, Val).

interpret(nth(N, Term), Val) :-
	!,
	interpret(N, I),
	interpret(Term, List),
	nth1(I, List, Val).

interpret(pos(Elem, Term), Val) :-
	!,	
	interpret(Elem, IElem),
	interpret(Term, List),
	nth1(Val, List, IElem).

% attribute 
interpret(rcp_att(Functor, [Arg]), R) :-
	!,
	interpret(Arg, rcp_rec(AssocList)),
	memberchk(Functor = Val, AssocList),
	interpret(Val, R).

% arithmetic
interpret(X + Y, R) :-
	interpret(X, Rx),
	interpret(Y, Ry),
	fd_aeq(R, Rx + Ry).

interpret(X - Y, R) :-
	!,
	interpret(X, Rx),
	interpret(Y, Ry),
	fd_aeq(R, Rx - Ry).

interpret(X * Y, R) :-
	!,
	interpret(X, Rx),
	interpret(Y, Ry),
	fd_aeq(R, Rx * Ry).

interpret(X / Y, R) :-
	!,
	interpret(X, Rx),
	interpret(Y, Ry),
	fd_aeq(R, Rx / Ry).

interpret(min(X, Y), R) :-
	!,
	interpret(X, Rx),
	interpret(Y, Ry),
	fd_aeq(R, min(Rx, Ry)).

interpret(max(X, Y), R) :-
	!,
	interpret(X, Rx),
	interpret(Y, Ry),
	fd_aeq(R, max(Rx, Ry)).

interpret(log(X, Y), R) :-
	!,
	interpret(X, B),
	interpret(Y, N),
	log(B, N, R).

interpret(exp(X, Y), R) :-
	!,
	interpret(X, B),
	interpret(Y, N),
	exp(B, N, R).

% comparison
interpret(X < Y, R) :-
	!,
	interpret(X, Rx),
	interpret(Y, Ry),
	fd_lt(Rx, Ry, R).

interpret(X =< Y, R) :-
	!,
	interpret(X, Rx),
	interpret(Y, Ry),
	fd_leq(Rx, Ry, R).

interpret(X = Y, R) :-
	!,
	interpret(X, Rx),
	interpret(Y, Ry),
	fd_eq(Rx, Ry, R).

interpret(X >= Y, R) :-
	!,
	interpret(X, Rx),
	interpret(Y, Ry),
	fd_geq(Rx, Ry, R).

interpret(X > Y, R) :-
	!,
	interpret(X, Rx),
	interpret(Y, Ry),
	fd_gt(Rx, Ry, R).

interpret(X # Y, R) :-
	!,
	interpret(X, Rx),
	interpret(Y, Ry),
	fd_neq(Rx, Ry, R).


% logic (normalisation)
interpret(1 and Q, R) :-
	!,
	interpret(Q, R).
interpret(P and 1, R) :-
	!,
	interpret(P, R).

interpret(0 or Q, R) :-
	!,
	interpret(Q, R).
interpret(P or 0, R) :-
	!,
	interpret(P, R).

interpret(not (P and Q), R) :-
	!,
	interpret(not P or not Q, R).
interpret(not (P or Q), R) :-
	!,
	interpret(not P and not Q, R).
interpret(not (not P), R) :-
	!,
	interpret(P, R).
interpret(not P, R) :-
	is_compop(P),
	!,
	neg_compop(P, NP),
	interpret(NP, R).

% logic (non-deterministic)
interpret(P and Q, R) :-
	srch_mode(_),
	!,
	leader(P and Q, Lead, Rest),
	fd_and(R1, R2, R),
				nl,
				write(seq_point(left-Lead)),nl,
	interpret(Lead, R1),
				%write(seq_point(right-Rest)),nl,
	interpret(Rest, R2).

interpret(P or Q, R) :-
	srch_mode(_),
	!,  
	leader(P or Q, Lead, Rest),
	(
				%write(choice_point(left-Lead)),nl,
	  interpret(Lead, R)
	;
				%write(choice_point(right-Rest)),nl,
	  interpret(Rest, R)
	).

interpret(P implies Q, R) :-
	srch_mode(_),
	!,
	interpret(P, RP),
				write(impl(RP, Q)),nl,
	(
	  fd_aeq(RP, 1),
	  interpret(Q, R)
	;
	  fd_aeq(RP, 0),
	  fd_aeq(R, 1)
	).


% logic (deterministic)
interpret(not P, R) :-
	!,
	fd_not(RP, R),
	interpret(P, RP).

interpret(P and Q, R) :-
	!,
	fd_and(RP, RQ, R),
	interpret(P, RP),
	interpret(Q, RQ).

interpret(P or Q, R) :-
	!,
	fd_or(RP, RQ, R),
	interpret(P, RP),
	interpret(Q, RQ).

interpret(P implies Q, R) :-
	!,
	fd_implies(RP, RQ, R),
	interpret(P, RP),
	interpret(Q, RQ).

interpret(P equiv Q, R) :-
	!,
	fd_equiv(RP, RQ, R),
	interpret(P, RP),
	interpret(Q, RQ).

% finite domain
interpret(domain(Term, L, U), 1) :-
	!,
	interpret(Term, VarList),
	interpret(L, Lb),
	interpret(U, Ub),
	domain(VarList, Lb, Ub).

% global constraints
interpret(all_different(Term), 1) :-
	!,
	interpret(Term, VarList),
	all_different(VarList).

interpret(cumulative(TermOs, TermDs, TermHs, TermL), 1) :-
	!,
	interpret(TermOs, Os),
	interpret(TermDs, Ds),	
	interpret(TermHs, Hs),
	interpret(TermL, L),
	length(Os, N),
	rcp_gen_variables(N, Vs),
	call(rcp_cumulative_tasks(Os, Ds, Hs, Vs, Tasks)),
	domain(Vs, 0, 1000),
	cumulative(Tasks, [limit(L), global(true)]),
	write(cumulative(Tasks, [limit(L), global(true)])),nl.

interpret(disjoint2(Xs, Ys, Ss), 1) :-
	!,
	interpret(TermXs, Xs),
	interpret(TermYs, Ys),	
	interpret(TermSs, Ss),
	rcp_disjoint2_data(Xs,Ys,Ss, Data),
	disjoint2(Data),
	write(disjoint2(Data)),nl.

% higher order 
interpret(foldr(X, L, Op, I, E), R) :-
	!,
	interpret(X, Var),
	interpret(L, List),
	interpret(I, Id),
				%write(fold_right(Var, List, Op, Id, E)),nl,nl,	
	expand_foldr(Var, List, Op, Id, E, ExpFoldr),
				%write(expand_fold_right(ExpFoldr)),nl,nl,
	interpret(ExpFoldr, R).

interpret(foldl(X, L, Op, I, E), R) :-
	!,
	interpret(X, Var),
	interpret(L, List),
	interpret(I, Id),
				%write(fold_left(Var, List, Op, Id, E)),nl,nl,	
	expand_foldl(Var, List, Op, Id, E, ExpFoldl),
				%write(expand_fold_left(ExpFoldl)),nl,nl,	
	interpret(ExpFoldl, R).

interpret(let(X, V, E), R) :-
	!,
	interpret(X, Var),
	interpret(V, Val),
	env_new(EmptyEnv),
	env_add(Var, EmptyEnv, Val, Env),	
	substitution(Env, E, '', _, SubstE),
				%write(let_interpret(SubstE)),nl,nl,
	interpret(SubstE, R).

% definition
interpret(Term, R) :-
	functor(Term, Functor, A),
	functor(Head, Functor, A),
	user:rcp_rule(Head, Body),
	!,
				%write(Term),nl,
				%write('-->'),nl,
	Term =.. [Functor | Args],
	Head =.. [Functor | FormalArgs],
	counter_init(local_vars),
	env_new(EmptyEnv),
	env_add_list(FormalArgs, EmptyEnv, Args, Env),
				%write(Body),nl,
	substitution(Env, Body, Term, _, SubstBody),
				%write(SubstBody),nl,nl,
	interpret(SubstBody, R).
				%,write(Term),write('-->'),write(R),nl,nl.

% heuristics (variable/value)
interpret(variable_ordering(CriteriaList), 1) :-	
	!,
	variable_ordering(CriteriaList).

interpret(value_ordering(CriteriaList), 1) :-	
	!,
	value_ordering(CriteriaList).

% heuristics (conjunct/disjunct)
interpret(conjunct_ordering(CriteriaList), 1) :-	
	!,
				write(conjunct_ordering(CriteriaList)),nl,
	conjunct_ordering(CriteriaList).

interpret(disjunct_ordering(CriteriaList), 1) :-	
	!,
				write(conjunct_ordering(CriteriaList)),nl,
	disjunct_ordering(CriteriaList).

% search
interpret(labeling([]), 1) :-
	!.
interpret(labeling(Term), R) :-
	!,
				%write(labeling(Term)),nl,	
	interpret(Term, VarList),
				%write(labeling(VarList)),nl,	
	leader(VarList, Lead, Rest),
	rcp_assign_var(Lead),
				%write(assign(Lead)),nl,
	interpret(labeling(Rest), R).

% optimisation
interpret(minimize(P, C), 1) :-
	!,
	interpret(C, Rc),
				write(minimize(P, C-Rc)),nl,	
	minimize((interpret(P, 1), labeling([up], [Rc])), Rc).

interpret(maximize(P, C), 1) :-
	!,
	interpret(C, Rc),
				write(maximize(P, C-Rc)),nl,
	maximize((interpret(P, 1), labeling([down], [Rc])), Rc).

% search mode 
interpret(search(Term), R) :-
	!,
	set_srch_mode(reordering),
				write(set_srch_mode),nl,
				%write(search(Term)),nl,
	interpret(Term, R),
	unset_srch_mode,
				write(unset_srch_mode(R)),nl
	.

interpret(flip(Term), R) :-
	!,
	interpret(Term, ITerm),
	flip(ITerm, R).

interpret(rcp_write(Term), 1) :-
	!,	
	interpret(Term, R),
	format("\"--- User Output ---\"\n ~w \n ", [Term]),
	format(" ~~>\n ~w \n\"--- End User Output ---\"\n", [R]).

% dynamic mode
interpret(dynamic(Term), R) :-
				write(dyn(Term)),nl,	
	interpret(Term, R).

% list interpretation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
interpret_list([], []).
interpret_list([H | T], [IH | IT]) :-
	interpret(H, IH),
	interpret_list(T, IT).

% record interpretation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
interpret_record([], []).
interpret_record([K = V | T], [K = IV | IT]) :-
	interpret(V, IV),
	interpret_record(T, IT).

% fold right expansion
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
expand_foldr(_, [], _, Id, _, Id).
expand_foldr(Var, [H | T], Op, Id, E, Expanded) :-
	subst(Var, H, E, SubstF),
	expand_foldr(Var, T, Op, Id, E, SubstR),
	composition(Op, SubstF, SubstR, Expanded).

% fold left expansion
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
expand_foldl(_, [], _, Id, _, Id).
expand_foldl(Var, [H | T], Op, Id, E, Expanded) :-
	subst(Var, H, E, SubstF),	
	composition(Op, Id, SubstF, R),
	expand_foldl(Var, T, Op, R, E, Expanded).

% ordering
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

leader((G implies P) and Q, Lead, Rest) :-
	rcp_heuristics(conjunct, ConjCriteriaList),
	interpret(G, R),
	R == 1,
	write(impll(G, R)),nl,
	!,
	leader(P and Q, Lead, Rest).

% and tree
%%%%%%%%%%%%%%%%%%%%
leader(P and Q, P, Q) :-
	rcp_heuristics(conjunct, []),
	!.

leader(P and Q, P, Q) :-
	rcp_heuristics(conjunct, ConjCriteriaList),
	!,
				%write(rcp_heuristics(conjunct, ConjCriteriaList)),nl,
	lex_vector(P, ConjCriteriaList, [P], LexComponents),
				 write(lex_vector(P, LexComponents)),nl.
				%write(Q),nl,nl.

% or tree
%%%%%%%%%%%%%%%%%%%%
leader(P or Q, P, Q) :-
	rcp_heuristics(disjunct, []),
	!.
leader(P or Q, P, Q) :-
	rcp_heuristics(disjunct, DisjCriteriaList),
	!.
				%write(rcp_heuristics(disjunct, DisjCriteriaList)),nl.

% var list
%%%%%%%%%%%%%%%%%%%%
leader([], none, []) :-
	!.
leader([H | T], H, T) :-
	rcp_heuristics(variable, []),
	rcp_heuristics(value, []),
	!.
leader(VarList, H, T) :-
	rcp_heuristics(variable, VarCriteriaList),
	rcp_variables(VarList, VarList0),
	delete_bounded_rcp_variables(VarList0, VarList1),
	select_variable(VarList1, VarCriteriaList, H),
	delete(VarList1, H, T).

% fd var assignment
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_assign_var(none) :-
	!.
rcp_assign_var(RcpVar) :-
	rcp_heuristics(value, ValCriteriaList),
	assign_variable(RcpVar, ValCriteriaList).

% interpreter mode flag
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
interpret_mode :-
	interpreter_mode.

set_interpret_mode :-
	retractall(interpreter_mode),
	assert(interpreter_mode).

unset_interpret_mode :-
	retractall(interpreter_mode).


% function composition
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
composition([F], A, B, R) :- 
	compound(F), 
	F =.. Fs,
	length(Fs, N),
	N =< 2,
	!,
	composition(Fs, A, B, R).
composition([F], A, B, R) :-
	!,
	R =.. [F, A, B].
composition([F | Fs], A, B, R) :-
	!,
	composition(Fs, A, B, R1),
	R =.. [F, R1].
composition(F, A, B, R) :-
	F =.. Fs,
	composition(Fs, A, B, R).

% disjoint2  
%%%%%%%%%%%%%%%%%%%%
rcp_disjoint2_data([], [], [], []).
rcp_disjoint2_data([X | Xs], [Y | Ys], [S | Ss], [r(X, S, Y, S) | Rs]) :-
	rcp_disjoint2_data(Xs, Ys, Ss, Rs).

% cumulative  
%%%%%%%%%%%%%%%%%%%%
rcp_cumulative_tasks([], [], [], [], []).
rcp_cumulative_tasks([O | Os], [D | Ds], [H | Hs], [E | Es], [task(O, D, E, H, 0) | Tasks]) :-
	rcp_cumulative_tasks(Os, Ds, Hs, Es, Tasks).

% geosr 
%%%%%%%%%%%%%%%%%%%%
traverse_presbox_shapes([],[]).
traverse_presbox_shapes([PreSboxShape | PreSboxShapes], SboxesShapes) :-
	traverse_presbox_shapes(PreSboxShapes,SboxShapes),
	traverse_presbox_boxes(PreSboxShape, Sboxes),
	append(Sboxes, SboxShapes, SboxesShapes).

traverse_presbox_boxes([],[]).
traverse_presbox_boxes([PreSboxBox | PreSboxBoxes],[Sbox | Sboxes]) :-
	traverse_presbox_boxes(PreSboxBoxes,Sboxes),
	PreSboxBox = [Sid, Offset, Sizes],
	Sbox =.. [sbox, Sid, Offset, Sizes].

% N fresh variables
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%g
rcp_gen_variables(0, []).
rcp_gen_variables(N, [_ | Vs]) :-	
	M is N - 1,
	rcp_gen_variables(M, Vs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% r2cp-specific constructs handling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% variables (global name-variable linking)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_var(Name, Value) :-	
	execution_state(break_level(0), private(Priv)),
	memberchk(r2cp(Priv0), Priv),
	memberchk((Name = Value), Priv0),
	(
 	  var(Value)
 	->
 	  put_atts(Value, rcp_var_name(Name))
 	;
 	  true
 	).
	
% variables 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_anavar(Value, Name) :-
	get_atts(Value, rcp_var_name(Name)).

% attribute (accessor)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_att(Att, rcp_rec(AssocList), Value) :-
	memberchk(Att = Value, AssocList).

% record (uid handling)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_rec(AssocList, rcp_rec(AssocList)) :-
	memberchk(uid = _Uid, AssocList),
	!.
rcp_rec(AssocList, rcp_rec(RTerm)) :-
	rcp_rec_uid_store(AssocList, Uid),
	append(AssocList, [uid = Uid], RTerm).

% range (to list)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_range([Lb .. Ub], List) :-
	!,
	gen_integers(Lb, Ub, List).
rcp_range(List, List) :-
	is_list(List).

% coumpound call
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_call(Term, R) :-
	Term =.. [F | Args],
	append(Args, [R], FArgs),
	TermToCall =.. [F | FArgs],
	call(TermToCall).

% call
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_assert(Term) :-
	var(Term),
	!,
	Term #= 1.
rcp_assert(Term) :-
	integer(Term),
	!,
	Term #= 1.
rcp_assert(Term) :-
	call(Term).

% variables (source definition's head)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
variables_bag(Expr) -->
	{
	 var(Expr)
	},
	!,
 	[Expr].
variables_bag(Expr) -->
	{
	 is_free_var(Expr),
	 interpret(Expr, Var)
	},
	!,
 	[Var].
variables_bag(Expr) -->
	{
	 atomic(Expr)
	},
	!.
variables_bag(Expr) -->
	{
	 functor(Expr, _, K)
	},
	variables_bag(0, K, Expr).

variables_bag(K, K, _) -->
	!.
variables_bag(I, K, Expr) -->
	{
	 J is I+1,
	 arg(J, Expr, Arg)
	},
	variables_bag(Arg),
	variables_bag(J, K, Expr).

% variables (source definition's head)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_variables_bag(Expr) -->
	{
	 var(Expr),	
	 rcp_anavar(Expr, N)	 
	},
	!,
 	[rcp_var(N, Expr)].
rcp_variables_bag(Expr) -->
	{
	 atomic(Expr)
	},
	!.
rcp_variables_bag(Expr) -->
	{
	 functor(Expr, _, K)
	},
	rcp_variables_bag(0, K, Expr).

rcp_variables_bag(K, K, _) -->
	!.
rcp_variables_bag(I, K, Expr) -->
	{
	 J is I+1,
	 arg(J, Expr, Arg)
	},
	rcp_variables_bag(Arg),
	rcp_variables_bag(J, K, Expr).

% multiset
%%%%%%%%%%%%%%%%%%%
variables(Expr, Bag) :-
	variables_bag(Expr, Bag, []).

rcp_variables(Expr, Bag) :-
	rcp_variables_bag(Expr, Bag, []).

% set
%%%%%%%%%%%%%%%%%%%
variables_set(Expr, Set) :-
	variables(Expr, Bag),
	list_to_set(Bag, Set).

rcp_variables_set(Expr, Set) :-
	rcp_variables(Expr, Bag),
	list_to_set(Bag, Set).


% debug printing
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_write(Term) :-
	format("\"--- User Output ---\"\n ~w \n\"--- End User Output ---\"\n\n",
	       [Term]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% r2cp builtins and FD operators handling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% arithmetic
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
max(A, B, R) :- call(R #= max(A, B)).

min(A, B, R) :- call(R #= min(A, B)).

log(A, B, R) :- R is ceiling(log(A, B)).

exp(A, B, R) :- R is ceiling(exp(A, B)).

% comparison (reified)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fd_lt(A, B, R) :- call(R #<=> A #< B).

fd_leq(A, B, R) :- call(R #<=> A #=< B).

fd_eq(A, B, R) :- call(R #<=> A #= B).

fd_geq(A, B, R) :- call(R #<=> A #>= B).

fd_gt(A, B, R) :- call(R #<=> A #> B).

fd_neq(A, B, R) :- call(R #<=> A #\= B).

% comparison (direct)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fd_lt(A, B) :- call(A #< B).

fd_leq(A, B) :- call(A #=< B).

fd_eq(A, B) :- call(A #= B).

fd_geq(A, B) :- call(A #>= B).

fd_gt(A, B) :- call(A #> B).

fd_neq(A, B) :- call(A #\= B).

% assignment
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fd_aeq(A, B) :- call(A #= B).

% logic connectives
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fd_and(P, Q, R) :- R #<=> (P #/\ Q).

fd_not(P, R) :- R #<=> (#\ P).

fd_or(P, Q, R) :- R #<=> (P #\/ Q).

fd_implies(P, Q, R) :- R #<=> (P #=> Q).

fd_equiv(P, Q, R) :- R #<=> (P #<=> Q).

% geosr global constraint
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_geost(Objects, ShiftedBoxes, ExternalConstraints) :-
	geosr(Objects,
	      ShiftedBoxes,
	      [(origin(O,D)---> O^x(D)),
	       (size(S,D)--->S^l(D)),
	       (end(O,D)--->
		   origin(O,D)+fold(S,sboxes([O^sid]),+,0,size(S,D))),
	       (overlap_sym(O1,O2,D)--->
		   end(O1,D)#>origin(O2,D)#/\end(O2,D)#>origin(O1,D)),
	       (overlap(O1,O2,Dims)--->
		   forall(D,Dims,overlap_sym(O1,O2,D))),
	       (non_overlapping(Items,Dims)--->
		   forall(O1,Items,
			  forall(O2,Items,
				 O1^oid#<O2^oid#=> #\overlap(O1,O2,Dims))))
	       | ExternalConstraints]).

rcp_geost(Objects, ShiftedBoxes, Options, ExternalConstraints) :-
	geost(Objects,
	      ShiftedBoxes,
	      Options,
	      [(origin(O,D)---> O^x(D)),
	       (size(S,D)--->S^l(D)),
	       (end(O,D)--->
		   origin(O,D)+fold(S,sboxes([O^sid]),+,0,size(S,D))),
	       (overlap_sym(O1,O2,D)--->
		   end(O1,D)#>origin(O2,D)#/\end(O2,D)#>origin(O1,D)),
	       (overlap(O1,O2,Dims)--->
		   forall(D,Dims,overlap_sym(O1,O2,D))),
	       (non_overlapping(Items,Dims)--->
		   forall(O1,Items,
			  forall(O2,Items,
				 O1^oid#<O2^oid#=> #\overlap(O1,O2,Dims))))
	       | ExternalConstraints]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% tree search and conjunct/disjunct heuristics handling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% conjunct ordering
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
conjunct_ordering(CriteriaList) :-
	retractall(rcp_heuristics(conjunct, _)),
	assert(rcp_heuristics(conjunct, CriteriaList)).	

% disjunct ordering
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
disjunct_ordering(CriteriaList) :-
	retractall(rcp_heuristics(disjunct, _)),
	assert(rcp_heuristics(disjunct, CriteriaList)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% labeling and variable/value heuristics handling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% variable ordering
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
variable_ordering(CriteriaList) :-
	retractall(rcp_heuristics(variable, _)),
	assert(rcp_heuristics(variable, CriteriaList)).	

% value ordering
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
value_ordering(CriteriaList) :-
	retractall(rcp_heuristics(value, _)),
	assert(rcp_heuristics(value, CriteriaList)).

% top-level labeling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_labeling([]) :-
	!.
rcp_labeling(VarList) :-
	rcp_heuristics(variable, VarCriteriaList),
	rcp_heuristics(value, ValCriteriaList),
	rcp_variables(VarList, RcpVarList1),
	delete_bounded_rcp_variables(RcpVarList1, RcpVarList),
	rcp_labeling(RcpVarList, VarCriteriaList, ValCriteriaList).

% heuritics labeling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_labeling([], _, _) :-
	!.
rcp_labeling(RcpVarList, VarCriteriaList, ValCriteriaList) :-
				%nl,nl,write('-labeling-\t'),
				%write(RcpVarList),nl,
	select_variable(RcpVarList, VarCriteriaList, BestRcpVar),
				%write('-selectingVar-\t'),
				%BestRcpVar = rcp_var(N, _),
				%write(N),nl,
	assign_variable(BestRcpVar, ValCriteriaList),	
	delete_bounded_rcp_variables(RcpVarList, UpdRcpVarList),
	rcp_labeling(UpdRcpVarList, VarCriteriaList, ValCriteriaList).

assign_variable(X, ValCriteriaList) :-
	var(X),
	!,
	rcp_anavar(X, rcp_var(N, Var)),
	select_value_enum(rcp_var(N, Var), ValCriteriaList, ValEnum),
	rcp_var(N, Var),
					%write('-selectingVals-\t'),
					%write(ValEnum-N-Var),nl,
	labeling(ValEnum, [Var]).
assign_variable(rcp_var(N, Var), ValCriteriaList) :-
	select_value_enum(rcp_var(N, Var), ValCriteriaList, ValEnum),
	rcp_var(N, Var),
					%write('-selectingVals-\t'),
					%write(ValEnum-N-Var),nl,
	labeling(ValEnum, [Var]).

select_value_enum(_, [], [up]) :-
	!.
select_value_enum(rcp_var(from(Name,N,I), Var), [ValCriterion | ValCriteriaList], ValEnum) :-
				%write('-valCriterion-\t'),
				%write(ValCriterion),nl,
				%write(rcp_var(from(Name,N,I), Var)),nl,
	ValCriterion =.. [Order, Term],
	Term =.. [F, NVarOrderCrit, Vars],
	TermToCall =.. [F, NVarOrderCrit, Name, Vars, Result],	
	on_exception(_Error,
		     call(TermToCall),
		     Result = undefined),
	select_value_enum(rcp_var(from(Name,N,I), Var), ValCriteriaList, ValEnum1),
	(
	  Result == Var
	->
	  append(ValEnum1, [Order], ValEnum)
	;
	  ValEnum = ValEnum1
	).

delete_bounded_rcp_variables([], []).
delete_bounded_rcp_variables([rcp_var(N, V) | T],  VT) :-
	rcp_var(N, V),
	integer(V),
	!,
	delete_bounded_rcp_variables(T, VT).
delete_bounded_rcp_variables([V | T], [V | VT]) :-
	delete_bounded_rcp_variables(T, VT).

% variable selection
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
select_variable([], _, none) :-
	!.
select_variable([BestVar | _], [], BestVar) :-
	!.
select_variable(RcpVarList, VarOrderCriteria, BestVar) :-	
	crit_order_list(VarOrderCriteria, CritOrderList),
				%write('-CritOrderList-\t\t'),
				%write(CritOrderList),nl,
	var_crit_vectors(RcpVarList, 1, VarOrderCriteria, VarCritValList),
				%write('-VarCritValList-\t\t'),
				%write(VarCritValList),nl,
	order_crit(1, CritOrderList, VarCritValList, [_-(_-BestVar)|_]).

crit_order_list([], []).
crit_order_list([LexVarOrderCrit | LexVarOrderCrits], [Order | OrderList]) :-
	LexVarOrderCrit =.. [Order | _],
	crit_order_list(LexVarOrderCrits, OrderList).

var_crit_vectors([], _, _, []).
var_crit_vectors([rcp_var(Name, Value) | Vars],
		 N,
		 LexVarOrderCriteria,
		 [LexTerm-(N-rcp_var(Name, Value)) | LexTerms]) :-
	var_crit_vector(Name, LexVarOrderCriteria, LexTerm),	
	M is N + 1,
	var_crit_vectors(Vars, M, LexVarOrderCriteria, LexTerms).

var_crit_vector(_, [], []).
var_crit_vector(from(Head, N, I),
		[LexVarOrderCriterion | LexVarOrderCriteria],
		[LexComponent | LexComponents]) :-
	interpret_mode,
	!,
	LexVarOrderCriterion =.. [_C, Term],
	subst(^, Head, Term, Subst),	
	on_exception(_Error,
		     interpret(Subst, LexComponent),
		     LexComponent = undefined),
	var_crit_vector(from(Head, N, I), LexVarOrderCriteria, LexComponents).
var_crit_vector(from(Head, N, I),
		[LexVarOrderCriterion | LexVarOrderCriteria],
		[LexComponent | LexComponents]) :-
	LexVarOrderCriterion =.. [_C, Term],
	Term =.. [F, NVarOrderCrit, Vars],
	TermToCall =.. [F, NVarOrderCrit, Head, Vars, LexComponent],
	on_exception(_Error,
		     call(TermToCall),
		     LexComponent = undefined),
	var_crit_vector(from(Head, N, I), LexVarOrderCriteria, LexComponents).

order_crit(N, OrderList, X, X) :-
	length(OrderList, M),
	N > M,	
	!.
order_crit(NCrit, OrderList, LexValGroup, OrderedLexVals) :-
				%write('-ncrit-\t\t'), write(NCrit),nl,
				%write('-lexico-\t'), write(LexValGroup),nl,
	select_criterion(LexValGroup, NCrit, Lex),
				%write('-lex-\t\t'), write(Lex),nl,
	nth1(NCrit, OrderList, Order),
				%write('-order-\t\t'), write(Order),nl,	
	keyclumps(Order, Lex, [OrderedLexClumped | _]),
				%write('-keyclumps-\t'),write(OrderedLexClumped),nl,
	order_full_lex(OrderedLexClumped, LexValGroup, OrderedFullLex),
				%write('-orderedFullLex-'), write(OrderedFullLex),nl,nl,
	MCrit is NCrit + 1,
	order_crit(MCrit, OrderList, OrderedFullLex, OrderedLexVals).

select_criterion([], _, []).
select_criterion([List-Tag | T], N, [Crit-Tag | Crits]) :-
	nth1(N, List, Crit),
	select_criterion(T, N, Crits).

order_full_lex([], _, []).
order_full_lex([_-T | Comps], LexValGroup, [C-T | FullComps]) :-
	order_full_lex(Comps, LexValGroup, FullComps),
	memberchk(C-T, LexValGroup).


% criteria values
%%%%%%%%%%%%%%%%%%%%
lex_vector(_P, [], _, []) :-
	!.
lex_vector([], _Crits, _CStack, []) :-
	!.
lex_vector([H | T], Crits, CStack, [LexComponentsH | LexComponents]) :-
	!,
	lex_vector(H, Crits, CStack, LexComponentsH),
	lex_vector(T, Crits, CStack, LexComponents).
lex_vector(P, [Crit | Crits], CStack, [LexComponent | LexComponents]) :-
	Crit =.. [_Order, Expr if is(^, MTerm)],
	functor(MTerm, Functor, Arity),
	functor(MTermVarArgs, Functor, Arity),
				%write(P-MTerm),nl,nl,
	(
	  MTermVarArgs = P
	->
	  MTerm =.. [Functor | FormalArgs],
	  P =.. [Functor | Args],	  
	  env_new(EmptyEnv),
	  env_add_list(FormalArgs, EmptyEnv, Args, Env),
	  substitution(Env, Expr, '', _, SubstExpr),
				%write(interpret(SubstExpr)),nl,
	  interpret(SubstExpr, LexComponent)
	;
	  LexComponent = undefined
	),
	lex_vector(P, Crits, CStack, LexComponents).


% key clumps
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
keyclumps(is, Lex, OrderedLexClumped) :-
	!,
	keyclumps_is(Lex, OrderedLexClumped).
keyclumps(Order, Lex, OrderedLexClumped) :-
	samsort(Order, Lex, OrderedLex),					
	keyclumps(OrderedLex, OrderedLexClumped).

keyclumps_is(Lex, OrderedLexClumped) :-
	keyclumps_is_(Lex, OrderedLexYes, OrderedLexNo),
	(
	  OrderedLexYes == []
	->
	  OrderedLexClumped = [OrderedLexNo]
	 ;
	  OrderedLexClumped = [OrderedLexYes, OrderedLexNo]
	).

keyclumps_is_([], [], []) :-
	!.
keyclumps_is_([W-(N-V) | T], Y, Nos) :-
	var(W),
	rcp_anavar(W, WName),
	rcp_var(WName, W) == V,
	!,
	keyclumps_is_(T, Yess, Nos),
	append([V-(N-V)], Yess, Y).
keyclumps_is_([H | T], Yess, N) :-
	keyclumps_is_(T, Yess, Nos),
	append([H], Nos,  N).	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% term comparison
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% greatest (variable ordering)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
greatest(_-_, undefined-_) :-
	!,
	fail.
greatest(undefined-X, _-Y) :-
	!,
	X @=< Y.	
greatest(T1-X, T2-Y) :-
	X @=< Y,
	T1 @>= T2.

% least (variable ordering)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
least(undefined-_, _-_) :-
	!,
	fail.
least(_-X, undefined-Y) :-
	!,
	X @=< Y.
least(T1-X, T2-Y) :-
	X @=< Y,
	T1 @=< T2.

% any (variable ordering)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
any(_-X, _-Y) :-
	X @=< Y.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% utilities
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% generate integer list
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
gen_integers(L, U, Ints) :- 
	gen_integers(L, U, L, Ints).

gen_integers(L, U, _, []) :-
	L > U,
	!.
gen_integers(_, U, U, [U]) :-
	!.
gen_integers(L, U, I, [I | Is]) :-
	N is I + 1,
	N =< U,
	N >= L,
	gen_integers(L, U, N, Is).

% copy_term (preserves all variables but V)
%%%%%%%%%%%%%%%%%%%%
copy_term_var(T, V, U) :-
	copy_term_var(T, V, _, _, U).

copy_term_var(T, V, Assoc, NewAssoc, U) :-
	var(T),
	T == V,
	!,
	(
	  empty_avl(Assoc)
	->
	  avl_store(V, Assoc, U, NewAssoc)
	;
	  avl_fetch(V, Assoc, U),
	  NewAssoc = Assoc
	).
copy_term_var(T, _, Assoc, Assoc, T) :-
	var(T),
	!.
copy_term_var(T, V, Assoc, NewAssoc, U) :-
	is_list(T),
	!,
	copy_term_var_list(T, V, Assoc, NewAssoc, U).
copy_term_var(T, V, Assoc, NewAssoc, U) :-
	compound(T),
	!,
	T =.. [F | Args],
	copy_term_var_list(Args, V, Assoc, NewAssoc, ArgsCopy),
	U =.. [F | ArgsCopy].
copy_term_var(T, _, Assoc, Assoc, T).
 
copy_term_var_list([], _V, Assoc, Assoc, []).
copy_term_var_list([T|Ts], V, Assoc, NewAssoc, [U|Us]) :-
	copy_term_var(T, V, Assoc, NewAssoc1, U),
	copy_term_var_list(Ts, V, NewAssoc1, NewAssoc, Us).

% variable instantiation 
%%%%%%%%%%%%%%%%%%%%
instantiate_variable(Var, Elem, Expr, InstExpr) :-
	copy_term_var(Var-Expr, Var, InstVar-InstExpr),
	InstVar = Elem.

% record uids handling
%%%%%%%%%%%%%%%%%%%%
init_rcp_rec_uid :-
	recorded(rcp_rec_uid(Record, Uid), rcp_rec_uid(Record, Uid), Ref),
	!,
	erase(Ref),
	init_rcp_rec_uid.
init_rcp_rec_uid :-
	counter_init(rcp_rec_uid_cpt).

rcp_rec_uid_store(Record0, Uid) :-
	rcp_name_vars_in_rec(Record0, Record),
	recorded(rcp_rec_uid(Record, Uid), rcp_rec_uid(Record, Uid), _),
	!.
rcp_rec_uid_store(Record0, Uid) :-	
	counter_inc(rcp_rec_uid_cpt, Uid),
	rcp_name_vars_in_rec(Record0, Record),
	recordz(rcp_rec_uid(Record, Uid), rcp_rec_uid(Record, Uid), _).

rcp_name_vars_in_rec([], []) :-
	!.
rcp_name_vars_in_rec([K = V0 | Rec0], [K = V | Rec]) :-
	var(V0),
	!,
	rcp_anavar(V0, V),
	rcp_name_vars_in_rec(Rec0, Rec).
rcp_name_vars_in_rec([K = V | Rec0], [K = V | Rec]) :-
	rcp_name_vars_in_rec(Rec0, Rec).

% flips Term arguments
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
flip_list([], []) :-
	!.
flip_list([X], [X]) :-
	!.
flip_list([X1, X2 | Xs], [X2, X1 | Ys]) :-
	flip_list(Xs, Ys).

rcp_flip(Term, FTerm) :- 
	is_list(Term),
	!,
	flip_list(Term, FTerm).
rcp_flip(Term, FTerm) :-
	compound(Term),
	Term =.. [Op | Args],
	flip_list(Args, FArgs),
	!,
	FTerm =.. [Op | FArgs].
rcp_flip(Term, Term).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% search mode tags {direct, reordering} handling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- dynamic search_mode/2.

srch_mode(Mode) :-
	search_mode(true, Mode).

set_srch_mode(Mode) :-
	nonvar(Mode), 
	retractall(search_mode(_, _)),
	assert(search_mode(true, Mode)).

unset_srch_mode :-
	retractall(search_mode(_, _)),
	assert(search_mode(false, none)).

neg_compop(P, NP) :-
	P =.. [F | Args],
	compop_neg(F, NF),
	NP =.. [NF | Args].

compop_neg(<, >=).
compop_neg(=<, >).
compop_neg(=, #).
compop_neg(>=, <).
compop_neg(>, =<).
compop_neg(#, =).

% is Term a comparison operator ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_compop(Term) :-
	functor(Term, Functor, Arity),
	lang(comp_ops, CompOps),
	memberchk(Functor/Arity, CompOps).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% higher-order functions (iterators)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% % ho macros
% %%%%%%%%%%%%%%%%%%%%
% forall(X, L, G, R) :-
% 				%write(forall(X,L,G)),nl,
% 	rcp_range(L, List),
% 	fold_left(X, List, ',', true, G, R).

% forall(X, L, G) :-
% 				%write(forall(X,L,G)),nl,
% 	rcp_range(L, List),
% 	fold_left(X, List, ',', true, G, _).

% exists(X, L, G, R) :-
% 				%write(exists(X,L,G)),nl,
% 	rcp_range(L, List),
% 	fold_left(X, List, ';', fail, G, R).

% exists(X, L, G) :-
% 				%write(exists(X,L,G)),nl,
% 	rcp_range(L, List),
% 	fold_left(X, List, ';', fail, G, _).

% let(X, E, G) :-
% 				%write(let(X,[E],G)),nl,
% 	E =.. [F|Args],
% 	append(Args, [R], RArgs),
% 	D =.. [F | RArgs],
% 	call(D),
% 	let(X, R, G, _).

% map(X, L, G, R) :-
% 				%write(map(X,L,G)),nl,
% 	rcp_range(L, List),
% 	fold_right(X, List, '.', [], G, R).

% % let
% %%%%%%%%%%%%%%%%%%%%
% let(V, X, G, R) :-
% 				%write(let(V, X, G)),nl,
% 	(
% 	  (
% 	    is_list(X)
% 	  ;
% 	    number(X)
% 	  )
% 	->
% 	  RX = X
% 	;
% 	  X =.. [F | XArgs],
% 	  append(XArgs, [RX], RXArgs),
% 	  Y =.. [F | RXArgs],
% 	  call(Y)
% 	),
% 	instantiate_variable(V, RX, G, Gc),
% 	Gc =.. [_ | Args],
% 	last(Args, R),
% 	call(Gc).

% % fold right
% %%%%%%%%%%%%%%%%%%%%
% foldr(X, L, F, I, G, R) :-
% 				%write(foldr(X, L, F, I, G)),nl,
% 	rcp_range(L, List),
% 	fold_right(X, List, F, I, G, R).

% fold_right(_, [],  _, I, _, I) :-
% 	!.
% fold_right(V, [X | Xs], F, I, G, R) :-
% 				%write(fold_right(V, X)),nl,
% 	fold_right(V, Xs, F, I, G, R1),
% 	instantiate_variable(V, X, G, Gc),	
% 	(
% 	  number(Gc)
% 	->
% 	  RGc = Gc
% 	;
% 	  Gc =.. [_H | GcArgs],
% 	  last(GcArgs, RGc),
% 	  instantiate_variable(RGc, RGc1, Gc, Gc1),
% 	  call(Gc1)
% 	),
% 	T =.. [F, RGc1, R1, R],
% 	call(T).

% % fold left
% %%%%%%%%%%%%%%%%%%%%
% foldl(X, L, F, I, G, R) :-
% 				%write(foldl(X, L, F, I, G)),nl,
% 	rcp_range(L, List),
% 	fold_left(X, List, F, I, G, R).

% fold_left(_, [], _, I, _, I) :-
% 	!.
% fold_left(V, [X | Xs], F, I, G, R) :-
% 				%write(fold_left(V, X)),nl,nl,	
% 	instantiate_variable(V, X, G, Gc),				
% 	(
% 	  number(Gc)
% 	->
% 	  RGc = Gc
% 	;
% 	  Gc =.. [_ | GcArgs],
% 	  last(GcArgs, RGc),
% 	  instantiate_variable(RGc, RGc1, Gc, Gc1),
% 	  call(Gc1)
% 	),	    
% 	T =.. [F, I, RGc1, R1],
% 	call(T),
% 	fold_left(V, Xs, F, R1, G, R).

% % loop until TG[V/X] is true,
% % G[V/X] otherwise
% %%%%%%%%%%%%%%%%%%%%
% until(V, X, _, TG) :-
% 	let(V, X, TG, _),
% 	!.
% until(V, X, G, TG) :-
% 	let(V, X, G, R),
% 	(
% 	  let(W,R,negative(W),_)
% 	->
% 	  true
% 	;
% 	  until(V, R, G, TG)
% 	).

% memberchk of assoc. values
% on term equality, not unification
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% mymemberchk(_, []) :-
% 	!.
% mymemberchk((Name = Value), [Name0 = Value0 | _]) :-
% 	Value == Value0,
% 	!,
% 	Name = Name0.
% mymemberchk((Name = Value), [_ = _ | Tail]) :-
% 	mymemberchk((Name = Value), Tail).

% rcp_anavar(Value, Name) :-
% 	execution_state(break_level(0), private(Priv)),
% 	memberchk(r2cp(Priv0), Priv),
% 	mymemberchk((Name = Value), Priv0).

/*

interval_split_foldl_1_search([], _, 1) :- 
	!.
interval_split_foldl_1_search([Cut_380|Tail], [L0,Max0,Min0,X0], _R_) :- 
	_A_=[X0,L0],
	(
	  var(X0)
	->
	  rcp_anavar(X0, XName0),
	  XName0 = from(item(rcp_rec([boxes=[rcp_rec([size=[XName,XName],uid=_])],positions=[[_,_]]])),_,_)
	;
	  XName = X0-ground
	),
	nth1(1, _A_, X),
	nth1(2, _A_, L),
	fd_min(X, XMin),
	XDLBL is XMin + L,
	(
	  write(leq(XName, XDLBL)),nl,
	  
	  fd_leq(X, XDLBL, _Leq_),
	  rcp_assert(_Leq_)
	;
	  write(gt(XName, XDLBL)),nl,
	  
	  fd_gt(X, XDLBL, _Gt_),
	  rcp_assert(_Gt_),		
	  interval_split_foldl_1_search(Tail, [L0,Max0,Min0,X0], _R_)
	).


rcp_interval_split_foldl_1_search([], _, 1) :- 
	!.
rcp_interval_split_foldl_1_search([Cut_718|Tail], [L,Max,Min,X], _R_) :- 
	List=[X,L],	
	nth1(1, List, X_721),
	nth1(2, List, L_724),
	fd_min(X_721, _C_),
	fd_aeq(_C_+L_724, XDLBL_726),
	(
	  fd_leq(X_721, XDLBL_726)
	;
	  fd_gt(X_721, XDLBL_726),
	  rcp_assert(E),
	  rcp_interval_split_foldl_1_search(Tail, [L,Max,Min,X], _R_)
	).

--

statistics(runtime, [T0| _]),

write(rect_try(Bin_3,Bin,Bin_2)),nl,

write('interval(Xs)'),nl,

write('interval(Ys)'),nl,
,
statistics(runtime, [T1|_]),
T is T1 - T0,
format('took ~3d sec.~n', [T])

N=23
took 4101.430 sec.
X = [4352,68,64] 
*/

% queens.pl 
% set_interpret_mode, interpret(solve_query, 1), interpret(free_var(q(1),0,1), Q1), interpret(free_var(q(2),0,1), Q2), interpret(free_var(q(3),0,1), Q3), interpret(free_var(q(4),0,1), Q4).

% psa.pl 
% set_interpret_mode, interpret(solve_query, 1), interpret([[origin(o2,1), origin(o2,2), origin(o2,3)], [origin(o3,1), origin(o3,2), origin(o3,3)], [origin(o4,1), origin(o4,2), origin(o4,3)]], Origins).

% rpacking.pl 
% set_interpret_mode, interpret(solve_query, 1), interpret(free_var(bin,0,1), W), interpret(free_var(bin,0,2), H), interpret(free_var(bin,0,3), A).

% bridge.pl 
% set_interpret_mode, interpret(solve_query, 1), interpret(free_var(last, 0, 1), Last).





% conj_order([1], greatest).
% conj_order([2], least).

% conj_crit(1, disj(A, B), R) :-
% 	!,
% 	rcp_att(duration, A, _A_),
% 	rcp_att(duration, B, _B_),
% 	fd_aeq(_A_+_B_, _R_).
	
% conj_crit(2, disj(A, B), R) :-
% 	fd_size(A, _A_),
% 	fd_size(B, _B_),
% 	fd_aeq(_A_+_B_, _R_).



% conj_crit(1, greatest, sum(map(Task, Tasks, duration(Task))), disjo(Tasks)).
% conj_crit(2, greatest, duration(A) + duration(B), disj(A, B)).
% conj_crit(3, least, domain_size(start(A)) + domain_size(start(B)), disj(A, B)).

% disj_crit(1, greates, duration(A), prec(A, B)).


% p_node(p1(T1, T2), end(T1) =< start(T2)).

% p_node(p2(T1, T2), end(T2) =< start(T1)).

% p_node(p(T1, T2),
% 	(
% 	  stable_sort_pair_lex_1([
% 				  [duration(T1)] - p1(T1, T2),
% 				  [duration(T2)] - p2(T1, T2)
% 				 ]),
% 	  disj_2(L)
% 	)	
%        ).

% p_node(q1, domain(cost, 0, 200)).

% p_node(q,
%        (
% 	 stable_sort_pair_lex_1([ [undef, undef, undef] - q1 |
% 			 map_flat_append(Tasks, resources_list,
% 					 map_flat_append(T1, Tasks,
% 							 map_flat_append(T2, Tasks,
% 									 uid(T1) < uid(T2) filter
% 									[
% 									 [undef,undef,undef] - pp(T1, T2),
									
% 									 [sum(map(Task, Tasks, duration(Task))),
% 									 duration(T1) + duration(T2),
% 									 domain_size(start(T1)) + domain_size(start(T2))] - p(T1, T2),

% 									 [undef,undef,undef] - ps(T1, T2)
									
% 									 ],
% 									 []),
% 							 []),
% 					 map_flat_append(N, [1 .. 10],
% 							 [ [undef, undef, undef] - pt(N) ], [])) ]
% 			 , L),
% 	 conj_2(L)
% 	 )
%       ).
	

% op_node(and,
% 	[
% 	 r_node([],
% 		domain([free_var(last,0,1)],0,200)
% 	       ),
	 
% 	 foldr(Tasks, resources_list, cons, [],
% 	       r_node([disjo(Tasks)],
% 		      foldr(T1, Tasks, cons, [],
% 			    foldr(T2, Tasks, cons, [],
% 				  uid(T1) < uid(T2) implies op_node(or,
% 								    [
% 								     r_node([prec(T1, T2), disj(T1, T2)],
% 									    start(T1) + duration(T1) =< start(T2)),
% 								     r_node([prec(T2, T1), disj(T1, T2)],
% 									    start(T2) + duration(T2) =< start(T1))
% 								    ]
% 								   )))))
% 	])



% rcp_disjunctives_foldr_2_search([], _, _, []) :- 
% 	!.
% rcp_disjunctives_foldr_2_search([H|Tail], [Tasks], [disjo(Tasks)] [_A_|_B_]) :-

% 	rcp_disjunctives_foldr_3_search(H, [Tasks], [disjo(Tasks)], _A_),
	
% 	rcp_disjunctives_foldr_2_search(Tail, [], [], _B_).

% rcp_disjunctives_foldr_1_search([], _, _, []) :- 
% 	!.
% rcp_disjunctives_foldr_1_search([H|Tail], [Tasks], [] [_A_|_B_]) :-

% 	rcp_disjunctives_foldr_2_search(H, [Tasks], [disjo(Tasks)], _A_),
	
% 	rcp_disjunctives_foldr_1_search(Tail, [], [], _B_).


% disjunctives_search :-
	
% 	eval_crit_list_conj(, L),
% 	sort_crit_list_conj(L, SL),
% 	it_goal_list_conj(SL)
	
% 	.


% resources_list = [[a1, a2, a3], [b1, b2, b3]].

% aot_node(and, root,
%      [
%       aot_node(and, root,
% 	       [
% 		domain([free_var(last,0,1)],0,200)
% 	       ]),

%       [
%        aot_node(and, disjo([a1, a2, a3]),
% 		[
% 		 [
% 		  uid(a1) < uid(a1) implies aot_node(or, disj(a1, a1),
% 						     [
% 						      aot_node(prec(a1, a1), start(a1) + duration(a1) =< start(a1)),
% 						      aot_node(prec(a1, a1), start(a1) + duration(a1) =< start(a1))
% 						     ]),
% 		  uid(a1) < uid(a2) implies aot_node(or, disj(a1, a2),
% 						     [
% 						      aot_node(prec(a1, a2), start(a1) + duration(a1) =< start(a2)),
% 						      aot_node(prec(a2, a1), start(a2) + duration(a2) =< start(a1))
% 						     ]),
% 		  uid(a1) < uid(a3) implies aot_node(or, disj(a1, a3),
% 						     [
% 						      aot_node(prec(a1, a3), start(a1) + duration(a1) =< start(a3)),
% 						      aot_node(prec(a3, a1), start(a3) + duration(a3) =< start(a1))
% 						     ])
% 		 ],
% 		 [
% 		  uid(a2) < uid(a1) implies aot_node(or, disj(a2, a1),
% 						     [
% 						      aot_node(prec(a2, a1), start(a2) + duration(a2) =< start(a1)),
% 						      aot_node(prec(a1, a2), start(a1) + duration(a1) =< start(a2))
% 						     ]),
% 		  uid(a2) < uid(a2) implies aot_node(or, disj(a2, a2),
% 						     [
% 						      aot_node(prec(a2, a2), start(a2) + duration(a2) =< start(a2)),
% 						      aot_node(prec(a2, a2), start(a2) + duration(a2) =< start(a2))
% 						     ]),
% 		  uid(a2) < uid(a3) implies aot_node(or, disj(a2, a3),
% 						     [
% 						      aot_node(prec(a2, a3), start(a2) + duration(a2) =< start(a3)),
% 						      aot_node(prec(a3, a2), start(a3) + duration(a3) =< start(a2))
% 						     ])
% 		 ],		 
% 		 [
% 		  uid(a3) < uid(a1) implies aot_node(or, disj(a3, a1),
% 						     [
% 						      aot_node(prec(a3, a1), start(a3) + duration(a3) =< start(a1)),
% 						      aot_node(prec(a1, a3), start(a1) + duration(a1) =< start(a3))
% 						     ]),
% 		  uid(a3) < uid(a2) implies aot_node(or, disj(a3, a2),
% 						     [
% 						      aot_node(prec(a3, a2), start(a3) + duration(a3) =< start(a2)),
% 						      aot_node(prec(a2, a3), start(a2) + duration(a2) =< start(a3))
% 						     ]),
% 		  uid(a3) < uid(a3) implies aot_node(or, disj(a3, a3),
% 						     [
% 						      aot_node(prec(a3, a3), start(a3) + duration(a3) =< start(a3)),
% 						      aot_node(prec(a3, a3), start(a3) + duration(a3) =< start(a3))
% 						     ])
% 		 ]
% 		]),
       
%        aot_node(and, disjo([b1, b2, b3]),
% 		foldr(T1, Tasks, cons, [],
% 		      foldr(T2, Tasks, cons, [],
% 			    uid(T1) < uid(T2) implies aot_node(or, disj(T1, T2),
% 							       [
% 								aot_node(prec(T1, T2), start(T1) + duration(T1) =< start(T2)),
% 								aot_node(prec(T2, T1), start(T2) + duration(T2) =< start(T1))
% 							       ])))
% 	       )
%       ]     
%      ])



%  node(disjunctives,
%       domain([free_var(last,0,1)],0,200))
%  and
%  foldr(Tasks, resources_list, and, true,
%        node(disjo(Tasks),      
% 	    foldr(T1, Tasks, and, true,
% 		  foldr(T2, Tasks, and, true, 
% 			uid(T1) < uid(T2) implies node(disj(T1, T2),					     
% 						       node(prec(T1, T2), start(T1) + duration(T1) =< start(T2))
% 						      or
% 						      node(prec(T2, T1), start(T2) + duration(T2) =< start(T1)))
% 		       ))))

